##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2013-2016 Uwe Hermann <uwe@hermann-uwe.de>
## Copyright (C) 2016 Chris Dreher <chrisdreher@hotmail.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from .lists import *

RX = 0
TX = 1

class Decoder(srd.Decoder):
    api_version = 3
    id = 'midi'
    name = 'MIDI'
    longname = 'Musical Instrument Digital Interface'
    desc = 'Musical Instrument Digital Interface (MIDI) protocol.'
    license = 'gplv2+'
    inputs = ['uart']
    outputs = []
    tags = ['Audio', 'PC']
    annotations = (
        ('text-verbose', 'Human-readable text (verbose)'),
        ('text-sysreal-verbose', 'Human-readable SysReal text (verbose)'),
        ('text-error', 'Human-readable Error text'),
    )
    annotation_rows = (
        ('normal', 'Normal', (0, 2)),
        ('sys-real', 'SysReal', (1,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.state = 'IDLE'
        self.status_byte = 0
        self.explicit_status_byte = False
        self.cmd = []
        self.ss = None
        self.es = None
        self.ss_block = None
        self.es_block = None

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        self.put(self.ss_block, self.es_block, self.out_ann, data)

    def get_note_name(self, channel, note):
        if channel != 10:
            return chromatic_notes[note]
        else:
            return 'assuming ' + percussion_notes.get(note, 'undefined')

    def check_for_garbage_flush(self, is_flushed):
        if is_flushed:
            if self.explicit_status_byte:
                self.cmd.insert(0, self.status_byte)
            self.handle_garbage_msg(None)

    def soft_clear_status_byte(self):
        self.explicit_status_byte = False

    def hard_clear_status_byte(self):
        self.status_byte = 0
        self.explicit_status_byte = False

    def set_status_byte(self, newbyte):
        self.status_byte = newbyte
        self.explicit_status_byte = True

    def handle_channel_msg_0x80(self, is_flushed):
        # Note off: 8n kk vv
        # n = channel, kk = note, vv = velocity
        c = self.cmd
        if len(c) < 2:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        note, velocity = c[0], c[1]
        note_name = self.get_note_name(chan, note)
        self.putx([0, ['Channel %d: %s (note = %d \'%s\', velocity = %d)' % \
                  (chan, status_bytes[msg][0], note, note_name, velocity),
                  'ch %d: %s %d, velocity = %d' % \
                  (chan, status_bytes[msg][1], note, velocity),
                  '%d: %s %d, vel %d' % \
                  (chan, status_bytes[msg][2], note, velocity)]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_0x90(self, is_flushed):
        # Note on: 9n kk vv
        # n = channel, kk = note, vv = velocity
        # If velocity == 0 that actually means 'note off', though.
        c = self.cmd
        if len(c) < 2:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        note, velocity = c[0], c[1]
        s = status_bytes[0x80] if (velocity == 0) else status_bytes[msg]
        note_name = self.get_note_name(chan, note)
        self.putx([0, ['Channel %d: %s (note = %d \'%s\', velocity = %d)' % \
                  (chan, s[0], note, note_name, velocity),
                  'ch %d: %s %d, velocity = %d' % \
                  (chan, s[1], note, velocity),
                  '%d: %s %d, vel %d' % \
                  (chan, s[2], note, velocity)]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_0xa0(self, is_flushed):
        # Polyphonic key pressure / aftertouch: An kk vv
        # n = channel, kk = polyphonic key pressure, vv = pressure value
        c = self.cmd
        if len(c) < 2:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        note, pressure = c[0], c[1]
        note_name = self.get_note_name(chan, note)
        self.putx([0, ['Channel %d: %s of %d for note = %d \'%s\'' % \
                  (chan, status_bytes[msg][0], pressure, note, note_name),
                  'ch %d: %s %d for note %d' % \
                  (chan, status_bytes[msg][1], pressure, note),
                  '%d: %s %d, N %d' % \
                  (chan, status_bytes[msg][2], pressure, note)]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_controller_0x44(self):
        # Legato footswitch: Bn 44 vv
        # n = channel, vv = value (<= 0x3f: normal, > 0x3f: legato)
        c = self.cmd
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        vv = c[1]
        t = ('normal', 'no') if vv <= 0x3f else ('legato', 'yes')
        self.putx([0, ['Channel %d: %s \'%s\' = %s' % \
                  (chan, status_bytes[msg][0],
                  control_functions[0x44][0], t[0]),
                  'ch %d: %s \'%s\' = %s' % \
                  (chan, status_bytes[msg][1],
                  control_functions[0x44][1], t[0]),
                  '%d: %s \'%s\' = %s' % \
                  (chan, status_bytes[msg][2],
                  control_functions[0x44][2], t[1])]])

    def handle_controller_0x54(self):
        # Portamento control (PTC): Bn 54 kk
        # n = channel, kk = source note for pitch reference
        c = self.cmd
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        kk = c[1]
        kk_name = self.get_note_name(chan, kk)
        self.putx([0, ['Channel %d: %s \'%s\' (source note = %d / %s)' % \
                  (chan, status_bytes[msg][0],
                  control_functions[0x54][0], kk, kk_name),
                  'ch %d: %s \'%s\' (source note = %d)' % \
                  (chan, status_bytes[msg][1],
                  control_functions[0x54][1], kk),
                  '%d: %s \'%s\' (src N %d)' % \
                  (chan, status_bytes[msg][2],
                  control_functions[0x54][2], kk)]])

    def handle_controller_generic(self):
        c = self.cmd
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        fn, param = c[0], c[1]
        default_name = 'undefined'
        ctrl_fn = control_functions.get(fn, default_name)
        if ctrl_fn == default_name:
            ctrl_fn = ('undefined 0x%02x' % fn, 'undef 0x%02x' % fn, '0x%02x' % fn)
        self.putx([0, ['Channel %d: %s \'%s\' (param = 0x%02x)' % \
                  (chan, status_bytes[msg][0], ctrl_fn[0], param),
                  'ch %d: %s \'%s\' (param = 0x%02x)' % \
                  (chan, status_bytes[msg][1], ctrl_fn[1], param),
                  '%d: %s \'%s\' is 0x%02x' % \
                  (chan, status_bytes[msg][2], ctrl_fn[2], param)]])

    def handle_channel_mode(self):
        # Channel Mode: Bn mm vv
        # n = channel, mm = mode number (120 - 127), vv = value
        c = self.cmd
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        mm, vv = c[0], c[1]
        mode_fn = control_functions.get(mm, ('undefined', 'undef', 'undef'))
        # Decode the value based on the mode number.
        vv_string = ('', '')
        if mm == 122:           # mode = local control?
            if vv == 0:
                vv_string = ('off', 'off')
            elif vv == 127:     # mode = poly mode on?
                vv_string = ('on', 'on')
            else:
                vv_string = ('(non-standard param value of 0x%02x)' % vv,
                            '0x%02x' % vv)
        elif mm == 126:         # mode = mono mode on?
            if vv != 0:
                vv_string = ('(%d channels)' % vv, '(%d ch)' % vv)
            else:
                vv_string = ('(channels \'basic\' through 16)',
                            '(ch \'basic\' thru 16)')
        elif vv != 0: # All other channel mode messages expect vv == 0.
            vv_string = ('(non-standard param value of 0x%02x)' % vv,
                        '0x%02x' % vv)
        self.putx([0, ['Channel %d: %s \'%s\' %s' % \
                      (chan, status_bytes[msg][0], mode_fn[0], vv_string[0]),
                      'ch %d: %s \'%s\' %s' % \
                      (chan, status_bytes[msg][1], mode_fn[1], vv_string[1]),
                      '%d: %s \'%s\' %s' % \
                      (chan, status_bytes[msg][2], mode_fn[2], vv_string[1])]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_0xb0(self, is_flushed):
        # Control change (or channel mode messages): Bn cc vv
        # n = channel, cc = control number (0 - 119), vv = control value
        c = self.cmd
        if len(c) < 2:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        if c[0] in range(0x78, 0x7f + 1):
            self.handle_channel_mode()
            return
        handle_ctrl = getattr(self, 'handle_controller_0x%02x' % c[0],
                              self.handle_controller_generic)
        handle_ctrl()
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_0xc0(self, is_flushed):
        # Program change: Cn pp
        # n = channel, pp = program number (0 - 127)
        c = self.cmd
        if len(c) < 1:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        pp = self.cmd[0] + 1
        change_type = 'instrument'
        name = ''
        if chan != 10:  # channel != percussion
            name = gm_instruments.get(pp, 'undefined')
        else:
            change_type = 'drum kit'
            name = drum_kit.get(pp, 'undefined')
        self.putx([0, ['Channel %d: %s to %s %d (assuming %s)' % \
            (chan, status_bytes[msg][0], change_type, pp, name),
            'ch %d: %s to %s %d' % \
            (chan, status_bytes[msg][1], change_type, pp),
            '%d: %s %d' % \
            (chan, status_bytes[msg][2], pp)]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_0xd0(self, is_flushed):
        # Channel pressure / aftertouch: Dn vv
        # n = channel, vv = pressure value
        c = self.cmd
        if len(c) < 1:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        vv = self.cmd[0]
        self.putx([0, ['Channel %d: %s %d' % (chan, status_bytes[msg][0], vv),
                      'ch %d: %s %d' % (chan, status_bytes[msg][1], vv),
                      '%d: %s %d' % (chan, status_bytes[msg][2], vv)]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_0xe0(self, is_flushed):
        # Pitch bend change: En ll mm
        # n = channel, ll = pitch bend change LSB, mm = pitch bend change MSB
        c = self.cmd
        if len(c) < 2:
            self.check_for_garbage_flush(is_flushed)
            return
        self.es_block = self.es
        msg, chan = self.status_byte & 0xf0, (self.status_byte & 0x0f) + 1
        ll, mm = self.cmd[0], self.cmd[1]
        decimal = (mm << 7) + ll
        self.putx([0, ['Channel %d: %s 0x%02x 0x%02x (%d)' % \
                      (chan, status_bytes[msg][0], ll, mm, decimal),
                      'ch %d: %s 0x%02x 0x%02x (%d)' % \
                      (chan, status_bytes[msg][1], ll, mm, decimal),
                      '%d: %s (%d)' % \
                      (chan, status_bytes[msg][2], decimal)]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg_generic(self, is_flushed):
        # TODO: It should not be possible to hit this code.
        # It currently can not be unit tested.
        msg_type = self.status_byte & 0xf0
        self.es_block = self.es
        self.putx([2, ['Unknown channel message type: 0x%02x' % msg_type]])
        self.cmd, self.state = [], 'IDLE'
        self.soft_clear_status_byte()

    def handle_channel_msg(self, newbyte):
        if newbyte is not None:
            if newbyte >= 0x80:
                self.set_status_byte(newbyte)
            else:
                self.cmd.append(newbyte)
        msg_type = self.status_byte & 0xf0
        handle_msg = getattr(self, 'handle_channel_msg_0x%02x' % msg_type,
                             self.handle_channel_msg_generic)
        handle_msg(newbyte is None)

    def handle_sysex_msg(self, newbyte):
        # SysEx message: 1 status byte, 1-3 manuf. bytes, x data bytes, EOX byte
        #
        # SysEx messages are variable length, can be terminated by EOX or
        # by any non-SysReal status byte, and it clears self.status_byte.
        #
        # Note: All System message codes don't utilize self.status_byte.
        self.hard_clear_status_byte()
        if newbyte != 0xf7 and newbyte is not None: # EOX
            self.cmd.append(newbyte)
            return
        self.es_block = self.es
        # Note: Unlike other methods, this code pops bytes out of self.cmd
        # to isolate the data.
        msg = self.cmd.pop(0)
        if len(self.cmd) < 1:
            self.putx([2, ['%s: truncated manufacturer code (<1 bytes)' % \
                          status_bytes[msg][0],
                          '%s: truncated manufacturer (<1 bytes)' % \
                          status_bytes[msg][1],
                          '%s: trunc. manu.' % status_bytes[msg][2]]])
            self.cmd, self.state = [], 'IDLE'
            return
        # Extract the manufacturer name (or SysEx realtime or non-realtime).
        m1 = self.cmd.pop(0)
        manu = (m1,)
        if m1 == 0x00:  # If byte == 0, then 2 more manufacturer bytes follow.
            if len(self.cmd) < 2:
                self.putx([2, ['%s: truncated manufacturer code (<3 bytes)' % \
                          status_bytes[msg][0],
                          '%s: truncated manufacturer (<3 bytes)' % \
                          status_bytes[msg][1],
                          '%s: trunc. manu.' % status_bytes[msg][2]]])
                self.cmd, self.state = [], 'IDLE'
                return
            manu = (m1, self.cmd.pop(0), self.cmd.pop(0))
        default_name = 'undefined'
        manu_name = sysex_manufacturer_ids.get(manu, default_name)
        if manu_name == default_name:
            if len(manu) == 3:
                manu_name = ('%s (0x%02x 0x%02x 0x%02x)' % \
                            (default_name, manu[0], manu[1], manu[2]),
                            default_name)
            else:
                manu_name = ('%s (0x%02x)' % (default_name, manu[0]),
                            default_name)
        else:
            manu_name = (manu_name, manu_name)
        # Extract the payload, display in 1 of 2 formats
        # TODO: Write methods to decode SysEx realtime & non-realtime payloads.
        payload0 = ''
        payload1 = ''
        while len(self.cmd) > 0:
            byte = self.cmd.pop(0)
            payload0 += '0x%02x ' % (byte)
            payload1 += '%02x ' % (byte)
        if payload0 == '':
            payload0 = '<empty>'
            payload1 = '<>'
        payload = (payload0, payload1)
        self.putx([0, ['%s: for \'%s\' with payload %s' % \
                      (status_bytes[msg][0], manu_name[0], payload[0]),
                      '%s: \'%s\', payload %s' % \
                      (status_bytes[msg][1], manu_name[1], payload[1]),
                      '%s: \'%s\', payload %s' % \
                      (status_bytes[msg][2], manu_name[1], payload[1])]])
        self.cmd, self.state = [], 'IDLE'

    def handle_syscommon_midi_time_code_quarter_frame_msg(self, newbyte):
        # MIDI time code quarter frame: F1 nd
        # n = message type
        # d = values
        #
        # Note: All System message codes don't utilize self.status_byte,
        # and System Exclusive and System Common clear it.
        c = self.cmd
        if len(c) < 2:
            if newbyte is None:
                self.handle_garbage_msg(None)
            return
        msg = c[0]
        nn, dd = (c[1] & 0x70) >> 4, c[1] & 0x0f
        group = ('System Common', 'SysCom', 'SC')
        self.es_block = self.es
        if nn != 7: # If message type does not contain SMPTE type.
            self.putx([0, ['%s: %s of %s, value 0x%01x' % \
                          (group[0], status_bytes[msg][0],
                          quarter_frame_type[nn][0], dd),
                          '%s: %s of %s, value 0x%01x' % \
                          (group[1], status_bytes[msg][1],
                          quarter_frame_type[nn][1], dd),
                          '%s: %s of %s, value 0x%01x' % \
                          (group[2], status_bytes[msg][2],
                          quarter_frame_type[nn][1], dd)]])
            self.cmd, self.state = [], 'IDLE'
            return
        tt = (dd & 0x6) >> 1
        self.putx([0, ['%s: %s of %s, value 0x%01x for %s' % \
                      (group[0], status_bytes[msg][0], \
                      quarter_frame_type[nn][0], dd, smpte_type[tt]),
                      '%s: %s of %s, value 0x%01x for %s' % \
                      (group[1], status_bytes[msg][1], \
                      quarter_frame_type[nn][1], dd, smpte_type[tt]),
                      '%s: %s of %s, value 0x%01x for %s' % \
                      (group[2], status_bytes[msg][2], \
                      quarter_frame_type[nn][1], dd, smpte_type[tt])]])
        self.cmd, self.state = [], 'IDLE'

    def handle_syscommon_msg(self, newbyte):
        # System common messages
        #
        # There are 5 simple formats (which are directly handled here) and
        # 1 complex one called MIDI time code quarter frame.
        #
        # Note: While the MIDI lists 0xf7 as a "system common" message, it
        # is actually only used with SysEx messages so it is processed there.
        #
        # Note: All System message codes don't utilize self.status_byte.
        self.hard_clear_status_byte()
        if newbyte is not None:
            self.cmd.append(newbyte)
        c = self.cmd
        msg = c[0]
        group = ('System Common', 'SysCom', 'SC')
        if msg == 0xf1:
            # MIDI time code quarter frame
            self.handle_syscommon_midi_time_code_quarter_frame_msg(newbyte)
            return
        elif msg == 0xf2:
            # Song position pointer: F2 ll mm
            # ll = LSB position, mm = MSB position
            if len(c) < 3:
                if newbyte is None:
                    self.handle_garbage_msg(None)
                return
            ll, mm = c[1], c[2]
            decimal = (mm << 7) + ll
            self.es_block = self.es
            self.putx([0, ['%s: %s 0x%02x 0x%02x (%d)' % \
                          (group[0], status_bytes[msg][0], ll, mm, decimal),
                          '%s: %s 0x%02x 0x%02x (%d)' % \
                          (group[1], status_bytes[msg][1], ll, mm, decimal),
                          '%s: %s (%d)' % \
                          (group[2], status_bytes[msg][2], decimal)]])
        elif msg == 0xf3:
            # Song select: F3 ss
            # ss = song selection number
            if len(c) < 2:
                if newbyte is None:
                    self.handle_garbage_msg(None)
                return
            ss = c[1]
            self.es_block = self.es
            self.putx([0, ['%s: %s number %d' % \
                          (group[0], status_bytes[msg][0], ss),
                          '%s: %s number %d' % \
                          (group[1], status_bytes[msg][1], ss),
                          '%s: %s # %d' % \
                          (group[2], status_bytes[msg][2], ss)]])
        elif msg == 0xf4 or msg == 0xf5 or msg == 0xf6:
            # Undefined 0xf4, Undefined 0xf5, and Tune Request (respectively).
            # All are only 1 byte long with no data bytes.
            self.es_block = self.es
            self.putx([0, ['%s: %s' % (group[0], status_bytes[msg][0]),
                          '%s: %s' % (group[1], status_bytes[msg][1]),
                          '%s: %s' % (group[2], status_bytes[msg][2])]])
        self.cmd, self.state = [], 'IDLE'

    def handle_sysrealtime_msg(self, newbyte):
        # System realtime message: 0b11111ttt (t = message type)
        #
        # Important: These messages are handled differently from all others
        # because they are allowed to temporarily interrupt other messages.
        # The interrupted messages resume after the realtime message is done.
        # Thus, they mostly leave 'self' the way it was found.
        #
        # Note: All System message codes don't utilize self.status_byte.
        old_ss_block, old_es_block = self.ss_block, self.es_block
        self.ss_block, self.es_block = self.ss, self.es
        group = ('System Realtime', 'SysReal', 'SR')
        self.putx([1, ['%s: %s' % (group[0], status_bytes[newbyte][0]),
                      '%s: %s' % (group[1], status_bytes[newbyte][1]),
                      '%s: %s' % (group[2], status_bytes[newbyte][2])]])
        self.ss_block, self.es_block = old_ss_block, old_es_block
        # Deliberately not resetting self.cmd or self.state.

    def handle_garbage_msg(self, newbyte):
        # Handle messages that are either not handled or are corrupt.
        self.es_block = self.es
        if newbyte is not None:
            self.cmd.append(newbyte)
            return
        payload = '<empty>'
        max_bytes = 16 # Put a limit on the length on the hex dump.
        for index in range(len(self.cmd)):
            if index == max_bytes:
                payload += ' ...'
                break
            if index == 0:
                payload = '0x%02x' % self.cmd[index]
            else:
                payload += ' 0x%02x' % self.cmd[index]
        self.putx([2, ['UNHANDLED DATA: %s' % payload,
                      'UNHANDLED', '???', '?']])
        self.cmd, self.state = [], 'IDLE'
        self.hard_clear_status_byte()

    def handle_state(self, state, newbyte):
        # 'newbyte' can either be:
        # 1. Value between 0x00-0xff, deal with the byte normally.
        # 2. Value of 'None' which means "flush any buffered data".
        if state == 'HANDLE CHANNEL MSG':
            self.handle_channel_msg(newbyte)
        elif state == 'HANDLE SYSEX MSG':
            self.handle_sysex_msg(newbyte)
        elif state == 'HANDLE SYSCOMMON MSG':
            self.handle_syscommon_msg(newbyte)
        elif state == 'HANDLE SYSREALTIME MSG':
            self.handle_sysrealtime_msg(newbyte)
        elif state == 'BUFFER GARBAGE MSG':
            self.handle_garbage_msg(newbyte)

    def get_next_state(self, newbyte):
        # 'newbyte' must be a valid byte between 0x00 and 0xff.
        #
        # Try to determine the state based off of the 'newbyte' parameter.
        if newbyte in range(0x80, 0xef + 1):
            return 'HANDLE CHANNEL MSG'
        if newbyte == 0xf0:
            return 'HANDLE SYSEX MSG'
        if newbyte in range(0xf1, 0xf7):
            return'HANDLE SYSCOMMON MSG'
        if newbyte in range(0xf8, 0xff + 1):
            return 'HANDLE SYSREALTIME MSG'
        # Passing 0xf7 is an error; messages don't start with 0xf7.
        if newbyte == 0xf7:
            return 'BUFFER GARBAGE MSG'
        # Next, base the state off of self.status_byte.
        if self.status_byte < 0x80:
            return 'BUFFER GARBAGE MSG'
        return self.get_next_state(self.status_byte)

    def decode(self, ss, es, data):
        ptype, rxtx, pdata = data
        state = 'IDLE'

        # For now, ignore all UART packets except the actual data packets.
        if ptype != 'DATA':
            return

        # We're only interested in the byte value (not individual bits).
        pdata = pdata[0]

        # Short MIDI overview:
        #  - Status bytes are 0x80-0xff, data bytes are 0x00-0x7f.
        #  - Most messages: 1 status byte, 1-2 data bytes.
        #  - Real-time system messages: always 1 byte.
        #  - SysEx messages: 1 status byte, n data bytes, EOX byte.
        #
        # Aspects of the MIDI protocol that complicate decoding:
        #  - MIDI System Realtime messages can briefly interrupt other
        #    messages already in progress.
        #  - "Running Status" allows for omitting the status byte in most
        #    scenarios if sequential messages have the same status byte.
        #  - System Exclusive (SysEx) messages can be terminated by ANY
        #    status byte (not limited to EOX byte).

        # State machine.
        if pdata >= 0x80 and pdata != 0xf7:
            state = self.get_next_state(pdata)
            if state != 'HANDLE SYSREALTIME MSG' and self.state != 'IDLE':
                # Flush the previous data since a new message is starting.
                self.handle_state(self.state, None)
            # Cache ss and es -after- flushing previous data.
            self.ss, self.es = ss, es
            # This is a status byte, remember the start sample.
            if state != 'HANDLE SYSREALTIME MSG':
                self.ss_block = ss
        elif self.state == 'IDLE' or self.state == 'BUFFER GARBAGE MSG':
            # Deal with "running status" or that we're buffering garbage.
            self.ss, self.es = ss, es
            if self.state == 'IDLE':
                self.ss_block = ss
            state = self.get_next_state(pdata)
        else:
            self.ss, self.es = ss, es
            state = self.state

        # Yes, this is intentionally _not_ an 'elif' here.
        if state != 'HANDLE SYSREALTIME MSG':
            self.state = state
        if state == 'BUFFER GARBAGE MSG':
            self.status_byte = 0
        self.handle_state(state, pdata)



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Marco Geisler <m-sigrok@mageis.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from collections import namedtuple
from .lists import *

ANN_STROBE, ANN_SINGLE_READ, ANN_SINGLE_WRITE, ANN_BURST_READ, \
    ANN_BURST_WRITE, ANN_STATUS_READ, ANN_STATUS, ANN_WARN = range(8)

Pos = namedtuple('Pos', ['ss', 'es'])
Data = namedtuple('Data', ['mosi', 'miso'])

class Decoder(srd.Decoder):
    api_version = 3
    id = 'cc1101'
    name = 'CC1101'
    longname = 'Texas Instruments CC1101'
    desc = 'Low-power sub-1GHz RF transceiver chip.'
    license = 'gplv2+'
    inputs = ['spi']
    outputs = []
    tags = ['IC', 'Wireless/RF']
    annotations = (
        ('strobe', 'Command strobe'),
        ('single_read', 'Single register read'),
        ('single_write', 'Single register write'),
        ('burst_read', 'Burst register read'),
        ('burst_write', 'Burst register write'),
        ('status', 'Status register'),
        ('warning', 'Warning'),
    )
    annotation_rows = (
        ('cmd', 'Commands', (ANN_STROBE,)),
        ('data', 'Data', (ANN_SINGLE_READ, ANN_SINGLE_WRITE, ANN_BURST_READ,
                            ANN_BURST_WRITE, ANN_STATUS_READ)),
        ('status', 'Status register', (ANN_STATUS,)),
        ('warnings', 'Warnings', (ANN_WARN,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.next()
        self.requirements_met = True
        self.cs_was_released = False

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def warn(self, pos, msg):
        '''Put a warning message 'msg' at 'pos'.'''
        self.put(pos.ss, pos.es, self.out_ann, [ANN_WARN, [msg]])

    def putp(self, pos, ann, msg):
        '''Put an annotation message 'msg' at 'pos'.'''
        self.put(pos.ss, pos.es, self.out_ann, [ann, [msg]])

    def putp2(self, pos, ann, msg1, msg2):
        '''Put an annotation message 'msg' at 'pos'.'''
        self.put(pos.ss, pos.es, self.out_ann, [ann, [msg1, msg2]])

    def next(self):
        '''Resets the decoder after a complete command was decoded.'''
        # 'True' for the first byte after CS# went low.
        self.first = True

        # The current command, and the minimum and maximum number
        # of data bytes to follow.
        self.cmd = None
        self.min = 0
        self.max = 0

        # Used to collect the bytes after the command byte
        # (and the start/end sample number).
        self.mb = []
        self.ss_mb = -1
        self.es_mb = -1

    def mosi_bytes(self):
        '''Returns the collected MOSI bytes of a multi byte command.'''
        return [b.mosi for b in self.mb]

    def miso_bytes(self):
        '''Returns the collected MISO bytes of a multi byte command.'''
        return [b.miso for b in self.mb]

    def decode_command(self, pos, b):
        '''Decodes the command byte 'b' at position 'pos' and prepares
        the decoding of the following data bytes.'''
        c = self.parse_command(b)
        if c is None:
            self.warn(pos, 'unknown command')
            return

        self.cmd, self.dat, self.min, self.max = c

        if self.cmd == 'Strobe':
            self.putp(pos, ANN_STROBE, self.format_command())
        else:
            # Don't output anything now, the command is merged with
            # the data bytes following it.
            self.ss_mb = pos.ss

    def format_command(self):
        '''Returns the label for the current command.'''
        if self.cmd in ('Read', 'Burst read', 'Write', 'Burst write', 'Status read'):
            return self.cmd
        if self.cmd == 'Strobe':
            reg = strobes.get(self.dat, 'unknown strobe')
            return '{} {}'.format(self.cmd, reg)
        else:
            return 'TODO Cmd {}'.format(self.cmd)

    def parse_command(self, b):
        '''Parses the command byte.

        Returns a tuple consisting of:
        - the name of the command
        - additional data needed to dissect the following bytes
        - minimum number of following bytes
        - maximum number of following bytes (None for infinite)
        '''

        addr = b & 0x3F
        if (addr < 0x30) or (addr == 0x3E) or (addr == 0x3F):
            if (b & 0xC0) == 0x00:
                return ('Write', addr, 1, 1)
            if (b & 0xC0) == 0x40:
                return ('Burst write', addr, 1, 99999)
            if (b & 0xC0) == 0x80:
                return ('Read', addr, 1, 1)
            if (b & 0xC0) == 0xC0:
                return ('Burst read', addr, 1, 99999)
            else:
                self.warn(pos, 'unknown address/command combination')
        else:
            if (b & 0x40) == 0x00:
                return ('Strobe', addr, 0, 0)
            if (b & 0xC0) == 0xC0:
                return ('Status read', addr, 1, 99999)
            else:
                self.warn(pos, 'unknown address/command combination')

    def decode_reg(self, pos, ann, regid, data):
        '''Decodes a register.

        pos   -- start and end sample numbers of the register
        ann   -- the annotation number that is used to output the register.
        regid -- may be either an integer used as a key for the 'regs'
                 dictionary, or a string directly containing a register name.'
        data  -- the register content.
        '''

        if type(regid) == int:
            # Get the name of the register.
            if regid not in regs:
                self.warn(pos, 'unknown register')
                return
            name = '{} ({:02X})'.format(regs[regid], regid)
        else:
            name = regid

        if regid == 'STATUS' and ann == ANN_STATUS:
            label = 'Status'
            self.decode_status_reg(pos, ann, data, label)
        else:
            if self.cmd in ('Write', 'Read', 'Status read', 'Burst read', 'Burst write'):
                label = '{}: {}'.format(self.format_command(), name)
            else:
                label = 'Reg ({}) {}'.format(self.cmd, name)
            self.decode_mb_data(pos, ann, data, label)

    def decode_status_reg(self, pos, ann, data, label):
        '''Decodes the data bytes 'data' of a status register at position
        'pos'. The decoded data is prefixed with 'label'.'''
        status = data[0]
        # bit 7 --> CHIP_RDYn
        if status & 0b10000000 == 0b10000000:
            longtext_chiprdy = 'CHIP_RDYn is high! '
        else:
            longtext_chiprdy = ''
        # bits 6:4 --> STATE
        state = (status & 0x70) >> 4
        longtext_state = 'STATE is {}, '.format(status_reg_states[state])
        # bits 3:0 --> FIFO_BYTES_AVAILABLE
        fifo_bytes = status & 0x0F
        if self.cmd in ('Single read', 'Status read', 'Burst read'):
            longtext_fifo = '{} bytes available in RX FIFO'.format(fifo_bytes)
        else:
            longtext_fifo = '{} bytes free in TX FIFO'.format(fifo_bytes)

        text = '{} = {:02X}'.format(label, status)
        longtext = ''.join([text, '; ', longtext_chiprdy, longtext_state, longtext_fifo])
        self.putp2(pos, ann, longtext, text)

    def decode_mb_data(self, pos, ann, data, label):
        '''Decodes the data bytes 'data' of a multibyte command at position
        'pos'. The decoded data is prefixed with 'label'.'''

        def escape(b):
            return '{:02X}'.format(b)

        data = ' '.join([escape(b) for b in data])
        text = '{} = {}'.format(label, data)
        self.putp(pos, ann, text)

    def finish_command(self, pos):
        '''Decodes the remaining data bytes at position 'pos'.'''

        if self.cmd == 'Write':
            self.decode_reg(pos, ANN_SINGLE_WRITE, self.dat, self.mosi_bytes())
        elif self.cmd == 'Burst write':
            self.decode_reg(pos, ANN_BURST_WRITE, self.dat, self.mosi_bytes())
        elif self.cmd == 'Read':
            self.decode_reg(pos, ANN_SINGLE_READ, self.dat, self.miso_bytes())
        elif self.cmd == 'Burst read':
            self.decode_reg(pos, ANN_BURST_READ, self.dat, self.miso_bytes())
        elif self.cmd == 'Strobe':
            self.decode_reg(pos, ANN_STROBE, self.dat, self.mosi_bytes())
        elif self.cmd == 'Status read':
            self.decode_reg(pos, ANN_STATUS_READ, self.dat, self.miso_bytes())
        else:
            self.warn(pos, 'unhandled command')

    def decode(self, ss, es, data):
        if not self.requirements_met:
            return

        ptype, data1, data2 = data

        if ptype == 'CS-CHANGE':
            if data1 is None:
                if data2 is None:
                    self.requirements_met = False
                    raise ChannelError('CS# pin required.')
                elif data2 == 1:
                    self.cs_was_released = True

            if data1 == 0 and data2 == 1:
                # Rising edge, the complete command is transmitted, process
                # the bytes that were sent after the command byte.
                if self.cmd:
                    # Check if we got the minimum number of data bytes
                    # after the command byte.
                    if len(self.mb) < self.min:
                        self.warn((ss, ss), 'missing data bytes')
                    elif self.mb:
                        self.finish_command(Pos(self.ss_mb, self.es_mb))

                self.next()
                self.cs_was_released = True

        elif ptype == 'DATA' and self.cs_was_released:
            mosi, miso = data1, data2
            pos = Pos(ss, es)

            if miso is None or mosi is None:
                self.requirements_met = False
                raise ChannelError('Both MISO and MOSI pins required.')

            if self.first:
                self.first = False
                # First MOSI byte is always the command.
                self.decode_command(pos, mosi)
                # First MISO byte is always the status register.
                self.decode_reg(pos, ANN_STATUS, 'STATUS', [miso])
            else:
                if not self.cmd or len(self.mb) >= self.max:
                    self.warn(pos, 'excess byte')
                else:
                    # Collect the bytes after the command byte.
                    if self.ss_mb == -1:
                        self.ss_mb = ss
                    self.es_mb = es
                    self.mb.append(Data(mosi, miso))



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2016 Elias Oenal <sigrok@eliasoenal.com>
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
## 1. Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
## 2. Redistributions in binary form must reproduce the above copyright notice,
##    this list of conditions and the following disclaimer in the documentation
##    and/or other materials provided with the distribution.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
## IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
## ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
## LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##

'''
The MDIO (Management Data Input/Output) protocol decoder supports the
MII Management serial bus (a bidirectional bus between the PHY and the STA),
with a clock line (MDC) and a bi-directional data line (MDIO).

MDIO is also known as SMI (Serial Management Interface).

It's part of the Ethernet standard.
'''

from .pd import Decoder



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2015 Jeremy Swanson <jeremy@rakocontrols.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from .lists import *

class SamplerateError(Exception):
    pass

class Decoder(srd.Decoder):
    api_version = 3
    id = 'dali'
    name = 'DALI'
    longname = 'Digital Addressable Lighting Interface'
    desc = 'Digital Addressable Lighting Interface (DALI) protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['Embedded/industrial', 'Lighting']
    channels = (
        {'id': 'dali', 'name': 'DALI', 'desc': 'DALI data line'},
    )
    options = (
        {'id': 'polarity', 'desc': 'Polarity', 'default': 'active-low',
            'values': ('active-low', 'active-high')},
    )
    annotations = (
        ('bit', 'Bit'),
        ('startbit', 'Startbit'),
        ('sbit', 'Select bit'),
        ('ybit', 'Individual or group'),
        ('address', 'Address'),
        ('command', 'Command'),
        ('reply', 'Reply data'),
        ('raw', 'Raw data'),
    )
    annotation_rows = (
        ('bits', 'Bits', (0,)),
        ('raw', 'Raw data', (7,)),
        ('fields', 'Fields', (1, 2, 3, 4, 5, 6)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.samplerate = None
        self.edges, self.bits, self.ss_es_bits = [], [], []
        self.state = 'IDLE'
        self.dev_type = None

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)
        self.old_dali = 1 if self.options['polarity'] == 'active-low' else 0

    def metadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value
            # One bit: 833.33us (one half low, one half high).
            # This is how may samples are in 1TE.
            self.halfbit = int((self.samplerate * 0.0008333) / 2.0)

    def putb(self, bit1, bit2, data):
        ss, es = self.ss_es_bits[bit1][0], self.ss_es_bits[bit2][1]
        self.put(ss, es, self.out_ann, data)

    def handle_bits(self, length):
        a, c, f, g, b = 0, 0, 0, 0, self.bits
        # Individual raw bits.
        for i in range(length):
            if i == 0:
                ss = max(0, self.bits[0][0])
            else:
                ss = self.ss_es_bits[i - 1][1]
            es = self.bits[i][0] + (self.halfbit * 2)
            self.ss_es_bits.append([ss, es])
            self.putb(i, i, [0, ['%d' % self.bits[i][1]]])
        # Bits[0:0]: Startbit
        s = ['Startbit: %d' % b[0][1], 'ST: %d' % b[0][1], 'ST', 'S', 'S']
        self.putb(0, 0, [1, s])
        self.putb(0, 0, [7, s])
        # Bits[1:8]
        for i in range(8):
            f |= (b[1 + i][1] << (7 - i))
        if length == 9: # BACKWARD Frame
            s = ['Reply: %02X' % f, 'Rply: %02X' % f,
                 'Rep: %02X' % f, 'R: %02X' % f, 'R']
            self.putb(1, 8, [7, s])
            s = ['Reply: %d' % f, 'Rply: %d' % f,
                 'Rep: %d' % f, 'R: %d' % f, 'R']
            self.putb(1, 8, [6, s])
            return

        # FORWARD FRAME
        # Bits[9:16]: Command/data (MSB-first)
        for i in range(8):
            c |= (b[9 + i][1] << (7 - i))
        # Raw output
        s = ['Raw data: %02X' % f, 'Raw: %02X' % f,
             'Raw: %02X' % f, 'R: %02X' % f, 'R']
        self.putb(1, 8, [7, s])
        s = ['Raw data: %02X' % c, 'Raw: %02X' % c,
                'Raw: %02X' % c, 'R: %02X' % c, 'R']
        self.putb(9, 16, [7, s])

        # Bits[8:8]: Select bit
        # s = ['Selectbit: %d' % b[8][1], 'SEL: %d' % b[8][1], 'SEL', 'SE', 'S']
        if b[8][1] == 1:
            s = ['Command', 'Comd', 'COM', 'CO', 'C']
        else:
            s = ['Arc Power Level', 'Arc Pwr', 'ARC', 'AC', 'A']
        self.putb(8, 8, [1, s])

        # f &= 254 # Clear the select bit.
        if f >= 254: # BROADCAST
            s = ['BROADCAST', 'Brdcast', 'BC', 'B', 'B']
            self.putb(1, 7, [5, s])
        elif f >= 160: # Extended command 0b10100000
            if f == 0xC1: # DALI_ENABLE_DEVICE_TYPE_X
                self.dev_type = -1
            x = extended_commands.get(f, ['Unknown', 'Unk'])
            s = ['Extended Command: %02X (%s)' % (f, x[0]),
                 'XC: %02X (%s)' % (f, x[1]),
                 'XC: %02X' % f, 'X: %02X' % f, 'X']
            self.putb(1, 8, [5, s])
        elif f >= 128: # Group
            # Bits[1:1]: Ybit
            s = ['YBit: %d' % b[1][1], 'YB: %d' % b[1][1], 'YB', 'Y', 'Y']
            self.putb(1, 1, [3, s])
            g = (f & 127) >> 1
            s = ['Group address: %d' % g, 'Group: %d' % g,
                'GP: %d' % g, 'G: %d' % g, 'G']
            self.putb(2,7, [4, s])
        else: # Short address
            # Bits[1:1]: Ybit
            s = ['YBit: %d' % b[1][1], 'YB: %d' % b[1][1], 'YB', 'Y', 'Y']
            self.putb(1, 1, [3, s])
            a = f >> 1
            s = ['Short address: %d' % a, 'Addr: %d' % a,
                'Addr: %d' % a, 'A: %d' % a, 'A']
            self.putb(2, 7, [4, s])

        # Bits[9:16]: Command/data (MSB-first)
        if f >= 160 and f < 254:
            if self.dev_type == -1:
                self.dev_type = c
                s = ['Type: %d' % c, 'Typ: %d' % c,
                     'Typ: %d' % c, 'T: %d' % c, 'D']
            else:
                self.dev_type = None
                s = ['Data: %d' % c, 'Dat: %d' % c,
                     'Dat: %d' % c, 'D: %d' % c, 'D']
        elif b[8][1] == 1:
            un = c & 0xF0
            ln = c & 0x0F
            if un == 0x10: # Set scene command
                x = ['Recall Scene %d' % ln, 'SC %d' % ln]
            elif un == 0x40:
                x = ['Store DTR as Scene %d' % ln, 'SC %d = DTR' % ln]
            elif un == 0x50:
                x = ['Delete Scene %d' % ln, 'DEL SC %d' % ln]
            elif un == 0x60:
                x = ['Add to Group %d' % ln, 'Grp %d Add' % ln]
            elif un == 0x70:
                x = ['Remove from Group %d' % ln, 'Grp %d Del' % ln]
            elif un == 0xB0:
                x = ['Query Scene %d Level' % ln, 'Sc %d Level' % ln]
            elif c >= 224: # Application specific commands
                if self.dev_type == 8:
                    x = dali_device_type8.get(c, ['Unknown App', 'Unk'])
                else:
                    x = ['Application Specific Command %d' % c, 'App Cmd %d' % c]
            else:
                x = dali_commands.get(c, ['Unknown', 'Unk'])
            s = ['Command: %d (%s)' % (c, x[0]), 'Com: %d (%s)' % (c, x[1]),
                 'Com: %d' % c, 'C: %d' % c, 'C']
        else:
            s = ['Arc Power Level: %d' % c, 'Level: %d' % c,
                 'Lev: %d' % c, 'L: %d' % c, 'L']
        self.putb(9, 16, [5, s])

    def reset_decoder_state(self):
        self.edges, self.bits, self.ss_es_bits = [], [], []
        self.state = 'IDLE'

    def decode(self):
        if not self.samplerate:
            raise SamplerateError('Cannot decode without samplerate.')
        bit = 0
        while True:
            # TODO: Come up with more appropriate self.wait() conditions.
            (dali,) = self.wait()
            if self.options['polarity'] == 'active-high':
                dali ^= 1 # Invert.

            # State machine.
            if self.state == 'IDLE':
                # Wait for any edge (rising or falling).
                if self.old_dali == dali:
                    continue
                self.edges.append(self.samplenum)
                self.state = 'PHASE0'
                self.old_dali = dali
                continue

            if self.old_dali != dali:
                self.edges.append(self.samplenum)
            elif self.samplenum == (self.edges[-1] + int(self.halfbit * 1.5)):
                self.edges.append(self.samplenum - int(self.halfbit * 0.5))
            else:
                continue

            bit = self.old_dali
            if self.state == 'PHASE0':
                self.phase0 = bit
                self.state = 'PHASE1'
            elif self.state == 'PHASE1':
                if (bit == 1) and (self.phase0 == 1): # Stop bit.
                    if len(self.bits) == 17 or len(self.bits) == 9:
                        # Forward or Backward.
                        self.handle_bits(len(self.bits))
                    self.reset_decoder_state() # Reset upon errors.
                    continue
                else:
                    self.bits.append([self.edges[-3], bit])
                    self.state = 'PHASE0'

            self.old_dali = dali



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2012 Uwe Hermann <uwe@hermann-uwe.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

class Decoder(srd.Decoder):
    api_version = 3
    id = 'i2cdemux'
    name = 'I²C demux'
    longname = 'I²C demultiplexer'
    desc = 'Demux I²C packets into per-slave-address streams.'
    license = 'gplv2+'
    inputs = ['i2c']
    outputs = [] # TODO: Only known at run-time.
    tags = ['Util']

    def __init__(self):
        self.reset()

    def reset(self):
        self.packets = [] # Local cache of I²C packets
        self.slaves = [] # List of known slave addresses
        self.stream = -1 # Current output stream
        self.streamcount = 0 # Number of created output streams

    def start(self):
        self.out_python = []

    # Grab I²C packets into a local cache, until an I²C STOP condition
    # packet comes along. At some point before that STOP condition, there
    # will have been an ADDRESS READ or ADDRESS WRITE which contains the
    # I²C address of the slave that the master wants to talk to.
    # We use this slave address to figure out which output stream should
    # get the whole chunk of packets (from START to STOP).
    def decode(self, ss, es, data):

        cmd, databyte = data

        # Add the I²C packet to our local cache.
        self.packets.append([ss, es, data])

        if cmd in ('ADDRESS READ', 'ADDRESS WRITE'):
            if databyte in self.slaves:
                self.stream = self.slaves.index(databyte)
                return

            # We're never seen this slave, add a new stream.
            self.slaves.append(databyte)
            self.out_python.append(self.register(srd.OUTPUT_PYTHON,
                                   proto_id='i2c-%s' % hex(databyte)))
            self.stream = self.streamcount
            self.streamcount += 1
        elif cmd == 'STOP':
            if self.stream == -1:
                raise Exception('Invalid stream!') # FIXME?

            # Send the whole chunk of I²C packets to the correct stream.
            for p in self.packets:
                self.put(p[0], p[1], self.out_python[self.stream], p[2])

            self.packets = []
            self.stream = -1
        else:
            pass # Do nothing, only add the I²C packet to our cache.



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2014 Gump Yang <gump.yang@gmail.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from .lists import *

class SamplerateError(Exception):
    pass

class Decoder(srd.Decoder):
    api_version = 3
    id = 'ir_nec'
    name = 'IR NEC'
    longname = 'IR NEC'
    desc = 'NEC infrared remote control protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['IR']
    channels = (
        {'id': 'ir', 'name': 'IR', 'desc': 'Data line'},
    )
    options = (
        {'id': 'polarity', 'desc': 'Polarity', 'default': 'active-low',
            'values': ('active-low', 'active-high')},
        {'id': 'cd_freq', 'desc': 'Carrier Frequency', 'default': 0},
    )
    annotations = (
        ('bit', 'Bit'),
        ('agc-pulse', 'AGC pulse'),
        ('longpause', 'Long pause'),
        ('shortpause', 'Short pause'),
        ('stop-bit', 'Stop bit'),
        ('leader-code', 'Leader code'),
        ('addr', 'Address'),
        ('addr-inv', 'Address#'),
        ('cmd', 'Command'),
        ('cmd-inv', 'Command#'),
        ('repeat-code', 'Repeat code'),
        ('remote', 'Remote'),
        ('warnings', 'Warnings'),
    )
    annotation_rows = (
        ('bits', 'Bits', (0, 1, 2, 3, 4)),
        ('fields', 'Fields', (5, 6, 7, 8, 9, 10)),
        ('remote', 'Remote', (11,)),
        ('warnings', 'Warnings', (12,)),
    )

    def putx(self, data):
        self.put(self.ss_start, self.samplenum, self.out_ann, data)

    def putb(self, data):
        self.put(self.ss_bit, self.samplenum, self.out_ann, data)

    def putd(self, data):
        name = self.state.title()
        d = {'ADDRESS': 6, 'ADDRESS#': 7, 'COMMAND': 8, 'COMMAND#': 9}
        s = {'ADDRESS': ['ADDR', 'A'], 'ADDRESS#': ['ADDR#', 'A#'],
             'COMMAND': ['CMD', 'C'], 'COMMAND#': ['CMD#', 'C#']}
        self.putx([d[self.state], ['%s: 0x%02X' % (name, data),
                  '%s: 0x%02X' % (s[self.state][0], data),
                  '%s: 0x%02X' % (s[self.state][1], data), s[self.state][1]]])

    def putstop(self, ss):
        self.put(ss, ss + self.stop, self.out_ann,
                 [4, ['Stop bit', 'Stop', 'St', 'S']])

    def putpause(self, p):
        self.put(self.ss_start, self.ss_other_edge, self.out_ann,
                 [1, ['AGC pulse', 'AGC', 'A']])
        idx = 2 if p == 'Long' else 3
        self.put(self.ss_other_edge, self.samplenum, self.out_ann,
                 [idx, [p + ' pause', '%s-pause' % p[0], '%sP' % p[0], 'P']])

    def putremote(self):
        dev = address.get(self.addr, 'Unknown device')
        buttons = command.get(self.addr, None)
        if buttons is None:
            btn = ['Unknown', 'Unk']
        else:
            btn = buttons.get(self.cmd, ['Unknown', 'Unk'])
        self.put(self.ss_remote, self.ss_bit + self.stop, self.out_ann,
                 [11, ['%s: %s' % (dev, btn[0]), '%s: %s' % (dev, btn[1]),
                 '%s' % btn[1]]])

    def __init__(self):
        self.reset()

    def reset(self):
        self.state = 'IDLE'
        self.ss_bit = self.ss_start = self.ss_other_edge = self.ss_remote = 0
        self.data = self.count = self.active = None
        self.addr = self.cmd = None

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def metadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value
        self.tolerance = 0.05 # +/-5%
        self.lc = int(self.samplerate * 0.0135) - 1 # 13.5ms
        self.rc = int(self.samplerate * 0.01125) - 1 # 11.25ms
        self.dazero = int(self.samplerate * 0.001125) - 1 # 1.125ms
        self.daone = int(self.samplerate * 0.00225) - 1 # 2.25ms
        self.stop = int(self.samplerate * 0.000652) - 1 # 0.652ms

    def compare_with_tolerance(self, measured, base):
        return (measured >= base * (1 - self.tolerance)
                and measured <= base * (1 + self.tolerance))

    def handle_bit(self, tick):
        ret = None
        if self.compare_with_tolerance(tick, self.dazero):
            ret = 0
        elif self.compare_with_tolerance(tick, self.daone):
            ret = 1
        if ret in (0, 1):
            self.putb([0, ['%d' % ret]])
            self.data |= (ret << self.count) # LSB-first
            self.count = self.count + 1
        self.ss_bit = self.samplenum

    def data_ok(self):
        ret, name = (self.data >> 8) & (self.data & 0xff), self.state.title()
        if self.count == 8:
            if self.state == 'ADDRESS':
                self.addr = self.data
            if self.state == 'COMMAND':
                self.cmd = self.data
            self.putd(self.data)
            self.ss_start = self.samplenum
            return True
        if ret == 0:
            self.putd(self.data >> 8)
        else:
            self.putx([12, ['%s error: 0x%04X' % (name, self.data)]])
        self.data = self.count = 0
        self.ss_bit = self.ss_start = self.samplenum
        return ret == 0

    def decode(self):
        if not self.samplerate:
            raise SamplerateError('Cannot decode without samplerate.')

        cd_count = None
        if self.options['cd_freq']:
            cd_count = int(self.samplerate / self.options['cd_freq']) + 1
        prev_ir = None

        self.active = 0 if self.options['polarity'] == 'active-low' else 1

        while True:
            # Detect changes in the presence of an active input signal.
            # The decoder can either be fed an already filtered RX signal
            # or optionally can detect the presence of a carrier. Periods
            # of inactivity (signal changes slower than the carrier freq,
            # if specified) pass on the most recently sampled level. This
            # approach works for filtered and unfiltered input alike, and
            # only slightly extends the active phase of input signals with
            # carriers included by one period of the carrier frequency.
            # IR based communication protocols can cope with this slight
            # inaccuracy just fine by design. Enabling carrier detection
            # on already filtered signals will keep the length of their
            # active period, but will shift their signal changes by one
            # carrier period before they get passed to decoding logic.
            if cd_count:
                (cur_ir,) = self.wait([{0: 'e'}, {'skip': cd_count}])
                if self.matched[0]:
                    cur_ir = self.active
                if cur_ir == prev_ir:
                    continue
                prev_ir = cur_ir
                self.ir = cur_ir
            else:
                (self.ir,) = self.wait({0: 'e'})

            if self.ir != self.active:
                # Save the non-active edge, then wait for the next edge.
                self.ss_other_edge = self.samplenum
                continue

            b = self.samplenum - self.ss_bit

            # State machine.
            if self.state == 'IDLE':
                if self.compare_with_tolerance(b, self.lc):
                    self.putpause('Long')
                    self.putx([5, ['Leader code', 'Leader', 'LC', 'L']])
                    self.ss_remote = self.ss_start
                    self.data = self.count = 0
                    self.state = 'ADDRESS'
                elif self.compare_with_tolerance(b, self.rc):
                    self.putpause('Short')
                    self.putstop(self.samplenum)
                    self.samplenum += self.stop
                    self.putx([10, ['Repeat code', 'Repeat', 'RC', 'R']])
                    self.data = self.count = 0
                self.ss_bit = self.ss_start = self.samplenum
            elif self.state == 'ADDRESS':
                self.handle_bit(b)
                if self.count == 8:
                    self.state = 'ADDRESS#' if self.data_ok() else 'IDLE'
            elif self.state == 'ADDRESS#':
                self.handle_bit(b)
                if self.count == 16:
                    self.state = 'COMMAND' if self.data_ok() else 'IDLE'
            elif self.state == 'COMMAND':
                self.handle_bit(b)
                if self.count == 8:
                    self.state = 'COMMAND#' if self.data_ok() else 'IDLE'
            elif self.state == 'COMMAND#':
                self.handle_bit(b)
                if self.count == 16:
                    self.state = 'STOP' if self.data_ok() else 'IDLE'
            elif self.state == 'STOP':
                self.putstop(self.ss_bit)
                self.putremote()
                self.ss_bit = self.ss_start = self.samplenum
                self.state = 'IDLE'



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2012 Uwe Hermann <uwe@hermann-uwe.de>
## Copyright (C) 2013 Matt Ranostay <mranostay@gmail.com>
## Copyright (C) 2014 alberink <alberink@stampfini.org>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

class Decoder(srd.Decoder):
    api_version = 3
    id = 'tca6408a'
    name = 'TI TCA6408A'
    longname = 'Texas Instruments TCA6408A'
    desc = 'Texas Instruments TCA6408A 8-bit I²C I/O expander.'
    license = 'gplv2+'
    inputs = ['i2c']
    outputs = []
    tags = ['Embedded/industrial', 'IC']
    annotations = (
        ('register', 'Register type'),
        ('value', 'Register value'),
        ('warnings', 'Warning messages'),
    )
    annotation_rows = (
        ('regs', 'Registers', (0, 1)),
        ('warnings', 'Warnings', (2,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.state = 'IDLE'
        self.chip = -1

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        self.put(self.ss, self.es, self.out_ann, data)

    def handle_reg_0x00(self, b):
        self.putx([1, ['State of inputs: %02X' % b]])

    def handle_reg_0x01(self, b):
        self.putx([1, ['Outputs set: %02X' % b ]])

    def handle_reg_0x02(self, b):
        self.putx([1, ['Polarity inverted: %02X' % b]])

    def handle_reg_0x03(self, b):
        self.putx([1, ['Configuration: %02X' % b]])

    def handle_write_reg(self, b):
        if b == 0:
            self.putx([0, ['Input port', 'In', 'I']])
        elif b == 1:
            self.putx([0, ['Output port', 'Out', 'O']])
        elif b == 2:
            self.putx([0, ['Polarity inversion register', 'Pol', 'P']])
        elif b == 3:
            self.putx([0, ['Configuration register', 'Conf', 'C']])

    def check_correct_chip(self, addr):
        if addr not in (0x20, 0x21):
            self.putx([2, ['Warning: I²C slave 0x%02X not a TCA6408A '
                           'compatible chip.' % addr]])
            self.state = 'IDLE'

    def decode(self, ss, es, data):
        cmd, databyte = data

        # Store the start/end samples of this I²C packet.
        self.ss, self.es = ss, es

        # State machine.
        if self.state == 'IDLE':
            # Wait for an I²C START condition.
            if cmd != 'START':
                return
            self.state = 'GET SLAVE ADDR'
        elif self.state == 'GET SLAVE ADDR':
            self.chip = databyte
            self.state = 'GET REG ADDR'
        elif self.state == 'GET REG ADDR':
            # Wait for a data write (master selects the slave register).
            if cmd in ('ADDRESS READ', 'ADDRESS WRITE'):
                self.check_correct_chip(databyte)
            if cmd != 'DATA WRITE':
                return
            self.reg = databyte
            self.handle_write_reg(self.reg)
            self.state = 'WRITE IO REGS'
        elif self.state == 'WRITE IO REGS':
            # If we see a Repeated Start here, the master wants to read.
            if cmd == 'START REPEAT':
                self.state = 'READ IO REGS'
                return
            # Otherwise: Get data bytes until a STOP condition occurs.
            if cmd == 'DATA WRITE':
                handle_reg = getattr(self, 'handle_reg_0x%02x' % self.reg)
                handle_reg(databyte)
            elif cmd == 'STOP':
                self.state = 'IDLE'
                self.chip = -1
        elif self.state == 'READ IO REGS':
            # Wait for an address read operation.
            if cmd == 'ADDRESS READ':
                self.state = 'READ IO REGS2'
                self.chip = databyte
                return
        elif self.state == 'READ IO REGS2':
            if cmd == 'DATA READ':
                handle_reg = getattr(self, 'handle_reg_0x%02x' % self.reg)
                handle_reg(databyte)
            elif cmd == 'STOP':
                self.state = 'IDLE'



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2015 Jeremy Swanson <jeremy@rakocontrols.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.
##

# DALI extended commands
extended_commands = {
    0xA1: ['Terminate special processes', 'Terminate'],
    0xA3: ['DTR = DATA', 'DTR'],
    0xA5: ['INITIALISE', 'INIT'],
    0xA7: ['RANDOMISE', 'RAND'],
    0xA9: ['COMPARE', 'COMP'],
    0xAB: ['WITHDRAW', 'WDRAW'],
    0xB1: ['SET SEARCH H', 'SAH'],
    0xB3: ['SET SEARCH M', 'SAM'],
    0xB5: ['SET SEARCH L', 'SAL'],
    0xB7: ['Program Short Address', 'ProgSA'],
    0xB9: ['Verify Short Address', 'VfySA'],
    0xBB: ['Query Short Address', 'QryShort'],
    0xBD: ['Physical Selection', 'PysSel'],
    0xC1: ['Enable Device Type X', 'EnTyp'],
    0xC3: ['DTR1 = DATA', 'DTR1'],
    0xC5: ['DTR2 = DATA', 'DTR2'],
    0xC7: ['Write Memory Location', 'WRI'],
}

# List of commands
dali_commands = {
    0x00: ['Immediate Off', 'IOFF'],
    0x01: ['Up 200ms', 'Up'],
    0x02: ['Down 200ms', 'Down'],
    0x03: ['Step Up', 'Step+'],
    0x04: ['Step Down', 'Step-'],
    0x05: ['Recall Maximum Level', 'Recall Max'],
    0x06: ['Recall Minimum Level', 'Recall Min'],
    0x07: ['Step down and off', 'Down Off'],
    0x08: ['Step ON and UP', 'On Up'],
    0x20: ['Reset', 'Rst'],
    0x21: ['Store Dim Level in DTR', 'Level -> DTR'],
    0x2A: ['Store DTR as Max Level', 'DTR->Max'],
    0x2B: ['Store DTR as Min Level', 'DTR->Min'],
    0x2C: ['Store DTR as Fail Level', 'DTR->Fail'],
    0x2D: ['Store DTR as Power On Level', 'DTR->Poweron'],
    0x2E: ['Store DTR as Fade Time', 'DTR->Fade'],
    0x2F: ['Store DTR as Fade Rate', 'DTR->Rate'],
    0x80: ['Store DTR as Short Address', 'DTR->Add'],
    0x81: ['Enable Memory Write', 'WEn'],
    0x90: ['Query Status', 'Status'],
    0x91: ['Query Ballast', 'Ballast'],
    0x92: ['Query Lamp Failure', 'LmpFail'],
    0x93: ['Query Power On', 'Power On'],
    0x94: ['Query Limit Error', 'Limit Err'],
    0x95: ['Query Reset', 'Reset State'],
    0x96: ['Query Missing Short Address', 'NoSrt'],
    0x97: ['Query Version', 'Ver'],
    0x98: ['Query DTR', 'GetDTR'],
    0x99: ['Query Device Type', 'Type'],
    0x9A: ['Query Physical Minimum', 'PhysMin'],
    0x9B: ['Query Power Fail', 'PowerFailed'],
    0x9C: ['Query DTR1', 'GetDTR1'],
    0x9D: ['Query DTR2', 'GetDTR2'],
    0xA0: ['Query Level', 'GetLevel'],
    0xA1: ['Query Max Level', 'GetMax'],
    0xA2: ['Query Min Level', 'GetMin'],
    0xA3: ['Query Power On', 'GetPwrOn'],
    0xA4: ['Query Fail Level', 'GetFail'],
    0xA5: ['Query Fade Rate', 'GetRate'],
    0xA6: ['Query Power Fail', 'PwrFail'],
    0xC0: ['Query Groups 0-7', 'GetGrpsL'],
    0xC1: ['Query Groups 7-15', 'GetGrpsH'],
    0xC2: ['Query BRNH', 'BRNH'],
    0xC3: ['Query BRNM', 'BRNM'],
    0xC4: ['Query BRNL', 'BRNL'],
    0xC5: ['Query Memory', 'GetMem'],
}

# DALI device type 8
dali_device_type8 = {
    0xE0: ['Set Temp X-Y Coordinate', 'Set X-Y'],
    0xE2: ['Activate Colour Set point', 'Activate SetPoint'],
    0xE7: ['Set Colour Temperature Tc', 'DTRs->ColTemp'],
    0xF9: ['Query Features', 'QryFeats'],
    0xFA: ['Query Current Setpoint Colour', 'GetSetPoint'],
}



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2017 Marcus Comstedt <marcus@mc.pp.se>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

step_wait_conds = (
    [{2: 'f'}, {0: 'l', 1: 'h'}],
    [{2: 'f'}, {0: 'h', 1: 'h'}, {1: 'l'}],
    [{2: 'f'}, {0: 'f'}, {1: 'l'}],
    [{2: 'f'}, {1: 'e'}],
)

class Decoder(srd.Decoder):
    api_version = 3
    id = 'iec'
    name = 'IEC'
    longname = 'Commodore bus'
    desc = 'Commodore serial IEEE-488 (IEC) bus protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = ['gpib']
    channels = (
        {'id': 'data', 'name': 'DATA', 'desc': 'Data I/O'},
        {'id': 'clk', 'name': 'CLK', 'desc': 'Clock'},
        {'id': 'atn', 'name': 'ATN', 'desc': 'Attention'},
    )
    optional_channels = (
        {'id': 'srq', 'name': 'SRQ', 'desc': 'Service request'},
    )
    annotations = (
        ('items', 'Items'),
        ('gpib', 'DAT/CMD'),
        ('eoi', 'EOI'),
    )
    annotation_rows = (
        ('bytes', 'Bytes', (0,)),
        ('gpib', 'DAT/CMD', (1,)),
        ('eoi', 'EOI', (2,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.saved_ATN = False
        self.saved_EOI = False
        self.ss_item = self.es_item = None
        self.step = 0
        self.bits = 0
        self.numbits = 0

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putb(self, data):
        self.put(self.ss_item, self.es_item, self.out_ann, data)

    def handle_bits(self):
        # Output the saved item.
        dbyte = self.bits
        dATN = self.saved_ATN
        dEOI = self.saved_EOI
        self.es_item = self.samplenum
        self.putb([0, ['%02X' % dbyte]])

        # Encode item byte to GPIB convention.
        self.strgpib = ' '
        if dATN: # ATN, decode commands.
            # Note: Commands < 0x20 are not used on IEC bus.
            if dbyte == 0x01: self.strgpib = 'GTL'
            if dbyte == 0x04: self.strgpib = 'SDC'
            if dbyte == 0x05: self.strgpib = 'PPC'
            if dbyte == 0x08: self.strgpib = 'GET'
            if dbyte == 0x09: self.strgpib = 'TCT'
            if dbyte == 0x11: self.strgpib = 'LLO'
            if dbyte == 0x14: self.strgpib = 'DCL'
            if dbyte == 0x15: self.strgpib = 'PPU'
            if dbyte == 0x18: self.strgpib = 'SPE'
            if dbyte == 0x19: self.strgpib = 'SPD'

            if dbyte == 0x3f: self.strgpib = 'UNL'
            if dbyte == 0x5f: self.strgpib = 'UNT'
            if dbyte > 0x1f and dbyte < 0x3f: # Address listener.
                self.strgpib = 'L' + chr(dbyte + 0x10)
            if dbyte > 0x3f and dbyte < 0x5f: # Address talker.
                self.strgpib = 'T' + chr(dbyte - 0x10)
            if dbyte > 0x5f and dbyte < 0x70: # Channel reopen.
                self.strgpib = 'R' + chr(dbyte - 0x30)
            if dbyte > 0xdf and dbyte < 0xf0: # Channel close.
                self.strgpib = 'C' + chr(dbyte - 0xb0)
            if dbyte > 0xef: # Channel open.
                self.strgpib = 'O' + chr(dbyte - 0xc0)
        else:
            if dbyte > 0x1f and dbyte < 0x7f:
                self.strgpib = chr(dbyte)
            if dbyte == 0x0a:
                self.strgpib = 'LF'
            if dbyte == 0x0d:
                self.strgpib = 'CR'

        self.putb([1, [self.strgpib]])
        self.strEOI = ' '
        if dEOI:
            self.strEOI = 'EOI'
        self.putb([2, [self.strEOI]])

    def decode(self):
        while True:

            data, clk, atn, srq = self.wait(step_wait_conds[self.step])

            if self.matched[0]:
                # Falling edge on ATN, reset step.
                self.step = 0

            if self.step == 0:
                # Don't use self.matched[1] here since we might come from
                # a step with different conds due to the code above.
                if data == 0 and clk == 1:
                    # Rising edge on CLK while DATA is low: Ready to send.
                    self.step = 1
            elif self.step == 1:
                if data == 1 and clk == 1:
                    # Rising edge on DATA while CLK is high: Ready for data.
                    self.ss_item = self.samplenum
                    self.saved_ATN = not atn
                    self.saved_EOI = False
                    self.bits = 0
                    self.numbits = 0
                    self.step = 2
                elif clk == 0:
                    # CLK low again, transfer aborted.
                    self.step = 0
            elif self.step == 2:
                if data == 0 and clk == 1:
                    # DATA goes low while CLK is still high, EOI confirmed.
                    self.saved_EOI = True
                elif clk == 0:
                    self.step = 3
            elif self.step == 3:
                if self.matched[1]:
                    if clk == 1:
                        # Rising edge on CLK; latch DATA.
                        self.bits |= data << self.numbits
                    elif clk == 0:
                        # Falling edge on CLK; end of bit.
                        self.numbits += 1
                        if self.numbits == 8:
                            self.handle_bits()
                            self.step = 0



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2016 fenugrec <fenugrec users.sourceforge.net>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

# TODO:
#  - Annotations are very crude and could be improved.
#  - Annotate every nibble? Would give insight on interrupted shifts.
#  - Annotate invalid "command" nibbles while SYNC==1?

import sigrokdecode as srd

class Decoder(srd.Decoder):
    api_version = 3
    id = 'aud'
    name = 'AUD'
    longname = 'Advanced User Debugger'
    desc = 'Renesas/Hitachi Advanced User Debugger (AUD) protocol.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['Debug/trace']
    channels = (
        {'id': 'audck', 'name': 'AUDCK', 'desc': 'AUD clock'},
        {'id': 'naudsync', 'name': 'nAUDSYNC', 'desc': 'AUD sync'},
        {'id': 'audata3', 'name': 'AUDATA3', 'desc': 'AUD data line 3'},
        {'id': 'audata2', 'name': 'AUDATA2', 'desc': 'AUD data line 2'},
        {'id': 'audata1', 'name': 'AUDATA1', 'desc': 'AUD data line 1'},
        {'id': 'audata0', 'name': 'AUDATA0', 'desc': 'AUD data line 0'},
    )
    annotations = (
        ('dest', 'Destination address'),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.ncnt = 0
        self.nmax = 0
        self.addr = 0
        self.lastaddr = 0
        self.ss = 0

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        self.put(self.ss, self.samplenum, self.out_ann, data)

    def handle_clk_edge(self, clk, sync, datapins):
        # Reconstruct nibble.
        nib = 0
        for i in range(4):
            nib |= datapins[3-i] << i

        # sync == 1: annotate if finished; update cmd.
        # TODO: Annotate idle level (nibble = 0x03 && SYNC=1).
        if sync == 1:
            if (self.ncnt == self.nmax) and (self.nmax != 0):
                # Done shifting an address: annotate.
                self.putx([0, ['0x%08X' % self.addr]])
                self.lastaddr = self.addr

            self.ncnt = 0
            self.addr = self.lastaddr
            self.ss = self.samplenum
            if nib == 0x08:
                self.nmax = 1
            elif nib == 0x09:
                self.nmax = 2
            elif nib == 0x0a:
                self.nmax = 4
            elif nib == 0x0b:
                self.nmax = 8
            else:
                # Undefined or idle.
                self.nmax = 0
        else:
            # sync == 0, valid cmd: start or continue shifting in nibbles.
            if (self.nmax > 0):
                # Clear tgt nibble.
                self.addr &= ~(0x0F << (self.ncnt * 4))
                # Set nibble.
                self.addr |= nib << (self.ncnt * 4)
                self.ncnt += 1

    def decode(self):
        while True:
            pins = self.wait({0: 'r'})
            clk = pins[0]
            sync = pins[1]
            d = pins[2:]
            self.handle_clk_edge(clk, sync, d)



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2019 Stephan Thiele <stephan.thiele@mailbox.org>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

'''
This decoder stacks on top of the 'spi' PD and decodes the button states
of an NES gamepad.

The SPI decoder needs to be configured as follows:

Clock polarity = 1
Clock phase    = 0
Bit order      = msb-first
Word size      = 8

Chip-select is not used and must not be assigned to any channel.

Hardware setup is as follows:
        ___
   GND |o  \
   CUP |o o| VCC
 OUT 0 |o o| D3
    D1 |o o| D4
       -----
NES Gamepad Connector

VCC   - Power 5V
GND   - Ground
CUP   - Shift register clock (CLK)
OUT 0 - Shift register latch (optional)
D1    - Gamepad data (MOSI)
D3    - Data (unused)
D4    - Data (unused)

Data pins D3 and D4 are not used by the standard gamepad but
by special controllers like the Nintento Zapper light gun.
'''

from .pd import Decoder



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2016 Daniel Schulte <trilader@schroedingers-bit.net>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd
from collections import namedtuple

class Ann:
    BIT, START, STOP, PARITY_OK, PARITY_ERR, DATA, WORD = range(7)

Bit = namedtuple('Bit', 'val ss es')

class Decoder(srd.Decoder):
    api_version = 3
    id = 'ps2'
    name = 'PS/2'
    longname = 'PS/2'
    desc = 'PS/2 keyboard/mouse interface.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['PC']
    channels = (
        {'id': 'clk', 'name': 'Clock', 'desc': 'Clock line'},
        {'id': 'data', 'name': 'Data', 'desc': 'Data line'},
    )
    annotations = (
        ('bit', 'Bit'),
        ('start-bit', 'Start bit'),
        ('stop-bit', 'Stop bit'),
        ('parity-ok', 'Parity OK bit'),
        ('parity-err', 'Parity error bit'),
        ('data-bit', 'Data bit'),
        ('word', 'Word'),
    )
    annotation_rows = (
        ('bits', 'Bits', (0,)),
        ('fields', 'Fields', (1, 2, 3, 4, 5, 6)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.bits = []
        self.bitcount = 0

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putb(self, bit, ann_idx):
        b = self.bits[bit]
        self.put(b.ss, b.es, self.out_ann, [ann_idx, [str(b.val)]])

    def putx(self, bit, ann):
        self.put(self.bits[bit].ss, self.bits[bit].es, self.out_ann, ann)

    def handle_bits(self, datapin):
        # Ignore non start condition bits (useful during keyboard init).
        if self.bitcount == 0 and datapin == 1:
            return

        # Store individual bits and their start/end samplenumbers.
        self.bits.append(Bit(datapin, self.samplenum, self.samplenum))

        # Fix up end sample numbers of the bits.
        if self.bitcount > 0:
            b = self.bits[self.bitcount - 1]
            self.bits[self.bitcount - 1] = Bit(b.val, b.ss, self.samplenum)
        if self.bitcount == 11:
            self.bitwidth = self.bits[1].es - self.bits[2].es
            b = self.bits[-1]
            self.bits[-1] = Bit(b.val, b.ss, b.es + self.bitwidth)

        # Find all 11 bits. Start + 8 data + odd parity + stop.
        if self.bitcount < 11:
            self.bitcount += 1
            return

        # Extract data word.
        word = 0
        for i in range(8):
            word |= (self.bits[i + 1].val << i)

        # Calculate parity.
        parity_ok = (bin(word).count('1') + self.bits[9].val) % 2 == 1

        # Emit annotations.
        for i in range(11):
            self.putb(i, Ann.BIT)
        self.putx(0, [Ann.START, ['Start bit', 'Start', 'S']])
        self.put(self.bits[1].ss, self.bits[8].es, self.out_ann, [Ann.WORD,
                 ['Data: %02x' % word, 'D: %02x' % word, '%02x' % word]])
        if parity_ok:
            self.putx(9, [Ann.PARITY_OK, ['Parity OK', 'Par OK', 'P']])
        else:
            self.putx(9, [Ann.PARITY_ERR, ['Parity error', 'Par err', 'PE']])
        self.putx(10, [Ann.STOP, ['Stop bit', 'Stop', 'St', 'T']])

        self.bits, self.bitcount = [], 0

    def decode(self):
        while True:
            # Sample data bits on falling clock edge.
            clock_pin, data_pin = self.wait({0: 'f'})
            self.handle_bits(data_pin)



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2018 Elias Oenal <sigrok@eliasoenal.com>
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
## 1. Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
## 2. Redistributions in binary form must reproduce the above copyright notice,
##    this list of conditions and the following disclaimer in the documentation
##    and/or other materials provided with the distribution.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
## IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
## ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
## LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
## CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
## SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
## INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
## CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
## ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
## POSSIBILITY OF SUCH DAMAGE.
##

import sigrokdecode as srd

MODULE_ID = {
    0x00: 'Unknown or unspecified',
    0x01: 'GBIC',
    0x02: 'Module/connector soldered to motherboard',
    0x03: 'SFP',
    0x04: '300 pin XSBI',
    0x05: 'XENPAK',
    0x06: 'XFP',
    0x07: 'XFF',
    0x08: 'XFP-E',
    0x09: 'XPAK',
    0x0a: 'X2',
    0x0B: 'DWDM-SFP',
    0x0C: 'QSFP',
    0x0D: 'QSFP+',
    0x0E: 'CFP',
    0x0F: 'CXP (TBD)',
    0x11: 'CFP2',
    0x12: 'CFP4',
}

class Decoder(srd.Decoder):
    api_version = 3
    id = 'cfp'
    name = 'CFP'
    longname = '100 Gigabit C form-factor pluggable'
    desc = '100 Gigabit C form-factor pluggable (CFP) protocol.'
    license = 'BSD'
    inputs = ['mdio']
    outputs = []
    tags = ['Networking']
    annotations = (
        ('register', 'Register'),
        ('decode', 'Decode'),
    )
    annotation_rows = (
        ('registers', 'Registers', (0,)),
        ('decodes', 'Decodes', (1,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        pass

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putx(self, data):
        self.put(self.ss, self.es, self.out_ann, data)

    def decode(self, ss, es, data):
        self.ss, self.es = ss, es
        for (clause45, clause45_addr, is_read, portad, devad, reg) in data:
            if not is_read:
                continue
            if clause45_addr in range(0x8000, 0x807F + 1):
                self.putx([0, ['CFP NVR 1: Basic ID register', 'NVR1']])
                if clause45_addr == 0x8000:
                    self.putx([1, ['Module identifier: %s' % \
                              MODULE_ID.get(reg, 'Reserved')]])
            elif clause45_addr in range(0x8080, 0x80FF + 1):
                self.putx([0, ['CFP NVR 2: Extended ID register', 'NVR2']])
            elif clause45_addr in range(0x8100, 0x817F + 1):
                self.putx([0, ['CFP NVR 3: Network lane specific register', 'NVR3']])
            elif clause45_addr in range(0x8180, 0x81FF + 1):
                self.putx([0, ['CFP NVR 4', 'NVR4']])
            elif clause45_addr in range(0x8400, 0x847F + 1):
                self.putx([0, ['Vendor NVR 1: Vendor data register', 'V-NVR1']])
            elif clause45_addr in range(0x8480, 0x84FF + 1):
                self.putx([0, ['Vendor NVR 2: Vendor data register', 'V-NVR2']])
            elif clause45_addr in range(0x8800, 0x887F + 1):
                self.putx([0, ['User NVR 1: User data register', 'U-NVR1']])
            elif clause45_addr in range(0x8880, 0x88FF + 1):
                self.putx([0, ['User NVR 2: User data register', 'U-NVR2']])
            elif clause45_addr in range(0xA000, 0xA07F + 1):
                self.putx([0, ['CFP Module VR 1: CFP Module level control and DDM register', 'Mod-VR1']])
            elif clause45_addr in range(0xA080, 0xA0FF + 1):
                self.putx([0, ['MLG VR 1: MLG Management Interface register', 'MLG-VR1']])



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2018 Stefan Brüns <stefan.bruens@rwth-aachen.de>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

PIN_DATA, PIN_RESET = range(2)
ROW_EDGE, ROW_WORD, ROW_RESET = range(3)

class Decoder(srd.Decoder):
    api_version = 3
    id = 'counter'
    name = 'Counter'
    longname = 'Edge counter'
    desc = 'Count the number of edges in a signal.'
    license = 'gplv2+'
    inputs = ['logic']
    outputs = []
    tags = ['Util']
    channels = (
        {'id': 'data', 'name': 'Data', 'desc': 'Data line'},
    )
    optional_channels = (
        {'id': 'reset', 'name': 'Reset', 'desc': 'Reset line'},
    )
    annotations = (
        ('edge_count', 'Edge count'),
        ('word_count', 'Word count'),
        ('word_reset', 'Word reset'),
    )
    annotation_rows = (
        ('edge_counts', 'Edges', (ROW_EDGE,)),
        ('word_counts', 'Words', (ROW_WORD,)),
        ('word_resets', 'Word resets', (ROW_RESET,)),
    )
    options = (
        {'id': 'data_edge', 'desc': 'Edges to count (data)', 'default': 'any',
            'values': ('any', 'rising', 'falling')},
        {'id': 'divider', 'desc': 'Count divider (word width)', 'default': 0},
        {'id': 'reset_edge', 'desc': 'Edge which clears counters (reset)',
            'default': 'falling', 'values': ('rising', 'falling')},
        {'id': 'edge_off', 'desc': 'Edge counter value after start/reset', 'default': 0},
        {'id': 'word_off', 'desc': 'Word counter value after start/reset', 'default': 0},
        {'id': 'dead_cycles', 'desc': 'Ignore this many edges after reset', 'default': 0},
        {'id': 'start_with_reset', 'desc': 'Assume decode starts with reset',
            'default': 'no', 'values': ('no', 'yes')},
    )

    def __init__(self):
        self.reset()

    def reset(self):
        pass

    def metadata(self, key, value):
        if key == srd.SRD_CONF_SAMPLERATE:
            self.samplerate = value

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def putc(self, cls, ss, annlist):
        self.put(ss, self.samplenum, self.out_ann, [cls, annlist])

    def decode(self):
        opt_edge_map = {'rising': 'r', 'falling': 'f', 'any': 'e'}

        data_edge = self.options['data_edge']
        divider = self.options['divider']
        if divider < 0:
            divider = 0
        reset_edge = self.options['reset_edge']

        condition = [{PIN_DATA: opt_edge_map[data_edge]}]
        have_reset = self.has_channel(PIN_RESET)
        if have_reset:
            cond_reset = len(condition)
            condition.append({PIN_RESET: opt_edge_map[reset_edge]})

        edge_count = int(self.options['edge_off'])
        edge_start = None
        word_count = int(self.options['word_off'])
        word_start = None

        if self.options['start_with_reset'] == 'yes':
            dead_count = int(self.options['dead_cycles'])
        else:
            dead_count = 0

        while True:
            self.wait(condition)
            now = self.samplenum

            if have_reset and self.matched[cond_reset]:
                edge_count = int(self.options['edge_off'])
                edge_start = now
                word_count = int(self.options['word_off'])
                word_start = now
                self.putc(ROW_RESET, now, ['Word reset', 'Reset', 'Rst', 'R'])
                dead_count = int(self.options['dead_cycles'])
                continue

            if dead_count:
                dead_count -= 1
                edge_start = now
                word_start = now
                continue

            # Implementation note: In the absence of a RESET condition
            # before the first data edge, any arbitrary choice of where
            # to start the annotation is valid. One may choose to emit a
            # narrow annotation (where ss=es), or assume that the cycle
            # which corresponds to the counter value started at sample
            # number 0. We decided to go with the latter here, to avoid
            # narrow annotations (see bug #1210). None of this matters in
            # the presence of a RESET condition in the input stream.
            if edge_start is None:
                edge_start = 0
            if word_start is None:
                word_start = 0

            edge_count += 1
            self.putc(ROW_EDGE, edge_start, ["{:d}".format(edge_count)])
            edge_start = now

            word_edge_count = edge_count - int(self.options['edge_off'])
            if divider and (word_edge_count % divider) == 0:
                word_count += 1
                self.putc(ROW_WORD, word_start, ["{:d}".format(word_count)])
                word_start = now



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2014 Jens Steinhauser <jens.steinhauser@gmail.com>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

class ChannelError(Exception):
    pass

regs = {
#   addr: ('name',        size)
    0x00: ('CONFIG',      1),
    0x01: ('EN_AA',       1),
    0x02: ('EN_RXADDR',   1),
    0x03: ('SETUP_AW',    1),
    0x04: ('SETUP_RETR',  1),
    0x05: ('RF_CH',       1),
    0x06: ('RF_SETUP',    1),
    0x07: ('STATUS',      1),
    0x08: ('OBSERVE_TX',  1),
    0x09: ('RPD',         1),
    0x0a: ('RX_ADDR_P0',  5),
    0x0b: ('RX_ADDR_P1',  5),
    0x0c: ('RX_ADDR_P2',  1),
    0x0d: ('RX_ADDR_P3',  1),
    0x0e: ('RX_ADDR_P4',  1),
    0x0f: ('RX_ADDR_P5',  1),
    0x10: ('TX_ADDR',     5),
    0x11: ('RX_PW_P0',    1),
    0x12: ('RX_PW_P1',    1),
    0x13: ('RX_PW_P2',    1),
    0x14: ('RX_PW_P3',    1),
    0x15: ('RX_PW_P4',    1),
    0x16: ('RX_PW_P5',    1),
    0x17: ('FIFO_STATUS', 1),
    0x1c: ('DYNPD',       1),
    0x1d: ('FEATURE',     1),
}

xn297_regs = {
    0x19: ('DEMOD_CAL',   5),
    0x1e: ('RF_CAL',      7),
    0x1f: ('BB_CAL',      5),
}

class Decoder(srd.Decoder):
    api_version = 3
    id = 'nrf24l01'
    name = 'nRF24L01(+)'
    longname = 'Nordic Semiconductor nRF24L01(+)'
    desc = '2.4GHz RF transceiver chip.'
    license = 'gplv2+'
    inputs = ['spi']
    outputs = []
    tags = ['IC', 'Wireless/RF']
    options = (
        {'id': 'chip', 'desc': 'Chip type',
            'default': 'nrf24l01', 'values': ('nrf24l01', 'xn297')},
    )
    annotations = (
        # Sent from the host to the chip.
        ('cmd', 'Commands sent to the device'),
        ('tx-data', 'Payload sent to the device'),

        # Returned by the chip.
        ('register', 'Registers read from the device'),
        ('rx-data', 'Payload read from the device'),

        ('warning', 'Warnings'),
    )
    ann_cmd = 0
    ann_tx = 1
    ann_reg = 2
    ann_rx = 3
    ann_warn = 4
    annotation_rows = (
        ('commands', 'Commands', (ann_cmd, ann_tx)),
        ('responses', 'Responses', (ann_reg, ann_rx)),
        ('warnings', 'Warnings', (ann_warn,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.next()
        self.requirements_met = True
        self.cs_was_released = False

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)
        if self.options['chip'] == 'xn297':
            regs.update(xn297_regs)

    def warn(self, pos, msg):
        '''Put a warning message 'msg' at 'pos'.'''
        self.put(pos[0], pos[1], self.out_ann, [self.ann_warn, [msg]])

    def putp(self, pos, ann, msg):
        '''Put an annotation message 'msg' at 'pos'.'''
        self.put(pos[0], pos[1], self.out_ann, [ann, [msg]])

    def next(self):
        '''Resets the decoder after a complete command was decoded.'''
        # 'True' for the first byte after CS went low.
        self.first = True

        # The current command, and the minimum and maximum number
        # of data bytes to follow.
        self.cmd = None
        self.min = 0
        self.max = 0

        # Used to collect the bytes after the command byte
        # (and the start/end sample number).
        self.mb = []
        self.mb_s = -1
        self.mb_e = -1

    def mosi_bytes(self):
        '''Returns the collected MOSI bytes of a multi byte command.'''
        return [b[0] for b in self.mb]

    def miso_bytes(self):
        '''Returns the collected MISO bytes of a multi byte command.'''
        return [b[1] for b in self.mb]

    def decode_command(self, pos, b):
        '''Decodes the command byte 'b' at position 'pos' and prepares
        the decoding of the following data bytes.'''
        c = self.parse_command(b)
        if c is None:
            self.warn(pos, 'unknown command')
            return

        self.cmd, self.dat, self.min, self.max = c

        if self.cmd in ('W_REGISTER', 'ACTIVATE'):
            # Don't output anything now, the command is merged with
            # the data bytes following it.
            self.mb_s = pos[0]
        else:
            self.putp(pos, self.ann_cmd, self.format_command())

    def format_command(self):
        '''Returns the label for the current command.'''
        if self.cmd == 'R_REGISTER':
            reg = regs[self.dat][0] if self.dat in regs else 'unknown register'
            return 'Cmd R_REGISTER "{}"'.format(reg)
        else:
            return 'Cmd {}'.format(self.cmd)

    def parse_command(self, b):
        '''Parses the command byte.

        Returns a tuple consisting of:
        - the name of the command
        - additional data needed to dissect the following bytes
        - minimum number of following bytes
        - maximum number of following bytes
        '''

        if (b & 0xe0) in (0b00000000, 0b00100000):
            c = 'R_REGISTER' if not (b & 0xe0) else 'W_REGISTER'
            d = b & 0x1f
            m = regs[d][1] if d in regs else 1
            return (c, d, 1, m)
        if b == 0b01010000:
            # nRF24L01 only
            return ('ACTIVATE', None, 1, 1)
        if b == 0b01100001:
            return ('R_RX_PAYLOAD', None, 1, 32)
        if b == 0b01100000:
            return ('R_RX_PL_WID', None, 1, 1)
        if b == 0b10100000:
            return ('W_TX_PAYLOAD', None, 1, 32)
        if b == 0b10110000:
            return ('W_TX_PAYLOAD_NOACK', None, 1, 32)
        if (b & 0xf8) == 0b10101000:
            return ('W_ACK_PAYLOAD', b & 0x07, 1, 32)
        if b == 0b11100001:
            return ('FLUSH_TX', None, 0, 0)
        if b == 0b11100010:
            return ('FLUSH_RX', None, 0, 0)
        if b == 0b11100011:
            return ('REUSE_TX_PL', None, 0, 0)
        if b == 0b11111111:
            return ('NOP', None, 0, 0)

    def decode_register(self, pos, ann, regid, data):
        '''Decodes a register.

        pos   -- start and end sample numbers of the register
        ann   -- is the annotation number that is used to output the register.
        regid -- may be either an integer used as a key for the 'regs'
                 dictionary, or a string directly containing a register name.'
        data  -- is the register content.
        '''

        if type(regid) == int:
            # Get the name of the register.
            if regid not in regs:
                self.warn(pos, 'unknown register')
                return
            name = regs[regid][0]
        else:
            name = regid

        # Multi byte register come LSByte first.
        data = reversed(data)

        if self.cmd == 'W_REGISTER' and ann == self.ann_cmd:
            # The 'W_REGISTER' command is merged with the following byte(s).
            label = '{}: {}'.format(self.format_command(), name)
        else:
            label = 'Reg {}'.format(name)

        self.decode_mb_data(pos, ann, data, label, True)

    def decode_mb_data(self, pos, ann, data, label, always_hex):
        '''Decodes the data bytes 'data' of a multibyte command at position
        'pos'. The decoded data is prefixed with 'label'. If 'always_hex' is
        True, all bytes are decoded as hex codes, otherwise only non
        printable characters are escaped.'''

        if always_hex:
            def escape(b):
                return '{:02X}'.format(b)
        else:
            def escape(b):
                c = chr(b)
                if not str.isprintable(c):
                    return '\\x{:02X}'.format(b)
                return c

        data = ''.join([escape(b) for b in data])
        text = '{} = "{}"'.format(label, data)
        self.putp(pos, ann, text)

    def finish_command(self, pos):
        '''Decodes the remaining data bytes at position 'pos'.'''

        if self.cmd == 'R_REGISTER':
            self.decode_register(pos, self.ann_reg,
                                 self.dat, self.miso_bytes())
        elif self.cmd == 'W_REGISTER':
            self.decode_register(pos, self.ann_cmd,
                                 self.dat, self.mosi_bytes())
        elif self.cmd == 'R_RX_PAYLOAD':
            self.decode_mb_data(pos, self.ann_rx,
                                self.miso_bytes(), 'RX payload', False)
        elif (self.cmd == 'W_TX_PAYLOAD' or
              self.cmd == 'W_TX_PAYLOAD_NOACK'):
            self.decode_mb_data(pos, self.ann_tx,
                                self.mosi_bytes(), 'TX payload', False)
        elif self.cmd == 'W_ACK_PAYLOAD':
            lbl = 'ACK payload for pipe {}'.format(self.dat)
            self.decode_mb_data(pos, self.ann_tx,
                                self.mosi_bytes(), lbl, False)
        elif self.cmd == 'R_RX_PL_WID':
            msg = 'Payload width = {}'.format(self.mb[0][1])
            self.putp(pos, self.ann_reg, msg)
        elif self.cmd == 'ACTIVATE':
            self.putp(pos, self.ann_cmd, self.format_command())
            if self.mosi_bytes()[0] != 0x73:
                self.warn(pos, 'wrong data for "ACTIVATE" command')

    def decode(self, ss, es, data):
        if not self.requirements_met:
            return

        ptype, data1, data2 = data

        if ptype == 'CS-CHANGE':
            if data1 is None:
                if data2 is None:
                    self.requirements_met = False
                    raise ChannelError('CS# pin required.')
                elif data2 == 1:
                    self.cs_was_released = True

            if data1 == 0 and data2 == 1:
                # Rising edge, the complete command is transmitted, process
                # the bytes that were send after the command byte.
                if self.cmd:
                    # Check if we got the minimum number of data bytes
                    # after the command byte.
                    if len(self.mb) < self.min:
                        self.warn((ss, ss), 'missing data bytes')
                    elif self.mb:
                        self.finish_command((self.mb_s, self.mb_e))

                self.next()
                self.cs_was_released = True
        elif ptype == 'DATA' and self.cs_was_released:
            mosi, miso = data1, data2
            pos = (ss, es)

            if miso is None or mosi is None:
                self.requirements_met = False
                raise ChannelError('Both MISO and MOSI pins required.')

            if self.first:
                self.first = False
                # First MOSI byte is always the command.
                self.decode_command(pos, mosi)
                # First MISO byte is always the status register.
                self.decode_register(pos, self.ann_reg, 'STATUS', [miso])
            else:
                if not self.cmd or len(self.mb) >= self.max:
                    self.warn(pos, 'excess byte')
                else:
                    # Collect the bytes after the command byte.
                    if self.mb_s == -1:
                        self.mb_s = ss
                    self.mb_e = es
                    self.mb.append((mosi, miso))



##
## This file is part of the libsigrokdecode project.
##
## Copyright (C) 2018 Michalis Pappas <mpappas@fastmail.fm>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, see <http://www.gnu.org/licenses/>.
##

import sigrokdecode as srd

WORD_ADDR_RESET         = 0x00
WORD_ADDR_SLEEP         = 0x01
WORD_ADDR_IDLE          = 0x02
WORD_ADDR_COMMAND       = 0x03

WORD_ADDR = {0x00: 'RESET', 0x01: 'SLEEP', 0x02: 'IDLE', 0x03: 'COMMAND'}

OPCODE_COUNTER          = 0x24
OPCODE_DERIVE_KEY       = 0x1c
OPCODE_DEV_REV          = 0x30
OPCODE_ECDH             = 0x43
OPCODE_GEN_DIG          = 0x15
OPCODE_GEN_KEY          = 0x40
OPCODE_HMAC             = 0x11
OPCODE_CHECK_MAC        = 0x28
OPCODE_LOCK             = 0x17
OPCODE_MAC              = 0x08
OPCODE_NONCE            = 0x16
OPCODE_PAUSE            = 0x01
OPCODE_PRIVWRITE        = 0x46
OPCODE_RANDOM           = 0x1b
OPCODE_READ             = 0x02
OPCODE_SHA              = 0x47
OPCODE_SIGN             = 0x41
OPCODE_UPDATE_EXTRA     = 0x20
OPCODE_VERIFY           = 0x45
OPCODE_WRITE            = 0x12

OPCODES = {
    0x01: 'Pause',
    0x02: 'Read',
    0x08: 'MAC',
    0x11: 'HMAC',
    0x12: 'Write',
    0x15: 'GenDig',
    0x16: 'Nonce',
    0x17: 'Lock',
    0x1b: 'Random',
    0x1c: 'DeriveKey',
    0x20: 'UpdateExtra',
    0x24: 'Counter',
    0x28: 'CheckMac',
    0x30: 'DevRev',
    0x40: 'GenKey',
    0x41: 'Sign',
    0x43: 'ECDH',
    0x45: 'Verify',
    0x46: 'PrivWrite',
    0x47: 'SHA',
}

ZONE_CONFIG             = 0x00
ZONE_OTP                = 0x01
ZONE_DATA               = 0x02

ZONES = {0x00: 'CONFIG', 0x01: 'OTP', 0x02: 'DATA'}

STATUS_SUCCESS          = 0x00
STATUS_CHECKMAC_FAIL    = 0x01
STATUS_PARSE_ERROR      = 0x03
STATUS_EXECUTION_ERROR  = 0x0f
STATUS_READY            = 0x11
STATUS_CRC_COMM_ERROR   = 0xff

STATUS = {
    0x00: 'Command success',
    0x01: 'Checkmac failure',
    0x03: 'Parse error',
    0x0f: 'Execution error',
    0x11: 'Ready',
    0xff: 'CRC / communications error',
}

class Decoder(srd.Decoder):
    api_version = 3
    id = 'atsha204a'
    name = 'ATSHA204A'
    longname = 'Microchip ATSHA204A'
    desc = 'Microchip ATSHA204A family crypto authentication protocol.'
    license = 'gplv2+'
    inputs = ['i2c']
    outputs = []
    tags = ['Security/crypto', 'IC', 'Memory']
    annotations = (
        ('waddr', 'Word address'),
        ('count', 'Count'),
        ('opcode', 'Opcode'),
        ('param1', 'Param1'),
        ('param2', 'Param2'),
        ('data', 'Data'),
        ('crc', 'CRC'),
        ('status', 'Status'),
        ('warning', 'Warning'),
    )
    annotation_rows = (
        ('frame', 'Frame', (0, 1, 2, 3, 4, 5, 6)),
        ('status', 'Status', (7,)),
        ('warnings', 'Warnings', (8,)),
    )

    def __init__(self):
        self.reset()

    def reset(self):
        self.state = 'IDLE'
        self.waddr = self.opcode = -1
        self.ss_block = self.es_block = 0
        self.bytes = []

    def start(self):
        self.out_ann = self.register(srd.OUTPUT_ANN)

    def output_tx_bytes(self):
        b = self.bytes
        if len(b) < 1: # Ignore wakeup.
            return
        self.waddr = b[0][2]
        self.put_waddr(b[0])
        if self.waddr == WORD_ADDR_COMMAND:
            count = b[1][2]
            self.put_count(b[1])
            if len(b) - 1 != count:
                self.put_warning(b[0][0], b[-1][1],
                    'Invalid frame length: Got {}, expecting {} '.format(
                    len(b) - 1, count))
                return
            self.opcode = b[2][2]
            self.put_opcode(b[2])
            self.put_param1(b[3])
            self.put_param2([b[4], b[5]])
            self.put_data(b[6:-2])
            self.put_crc([b[-2], b[-1]])

    def output_rx_bytes(self):
        b = self.bytes
        count = b[0][2]
        self.put_count(b[0])
        if self.waddr == WORD_ADDR_RESET:
            self.put_data([b[1]])
            self.put_crc([b[2], b[3]])
            self.put_status(b[0][0], b[-1][1], b[1][2])
        elif self.waddr == WORD_ADDR_COMMAND:
            if count == 4: # Status / Error.
                self.put_data([b[1]])
                self.put_crc([b[2], b[3]])
                self.put_status(b[0][0], b[-1][1], b[1][2])
            else:
                self.put_data(b[1:-2])
                self.put_crc([b[-2], b[-1]])

    def putx(self, s, data):
        self.put(s[0], s[1], self.out_ann, data)

    def puty(self, s, data):
        self.put(s[0][0], s[1][1], self.out_ann, data)

    def putz(self, ss, es, data):
        self.put(ss, es, self.out_ann, data)

    def put_waddr(self, s):
        self.putx(s, [0, ['Word addr: %s' % WORD_ADDR[s[2]]]])

    def put_count(self, s):
        self.putx(s, [1, ['Count: %s' % s[2]]])

    def put_opcode(self, s):
        self.putx(s, [2, ['Opcode: %s' % OPCODES[s[2]]]])

    def put_param1(self, s):
        op = self.opcode
        if op in (OPCODE_CHECK_MAC, OPCODE_COUNTER, OPCODE_DEV_REV,     \
                  OPCODE_ECDH, OPCODE_GEN_KEY, OPCODE_HMAC, OPCODE_MAC, \
                  OPCODE_NONCE, OPCODE_RANDOM, OPCODE_SHA, OPCODE_SIGN, \
                  OPCODE_VERIFY):
            self.putx(s, [3, ['Mode: %02X' % s[2]]])
        elif op == OPCODE_DERIVE_KEY:
            self.putx(s, [3, ['Random: %s' % s[2]]])
        elif op == OPCODE_PRIVWRITE:
            self.putx(s, [3, ['Encrypted: {}'.format('Yes' if s[2] & 0x40 else 'No')]])
        elif op == OPCODE_GEN_DIG:
            self.putx(s, [3, ['Zone: %s' % ZONES[s[2]]]])
        elif op == OPCODE_LOCK:
            self.putx(s, [3, ['Zone: {}, Summary: {}'.format(
                'DATA/OTP' if s[2] else 'CONFIG',
                'Ignored' if s[2] & 0x80 else 'Used')]])
        elif op == OPCODE_PAUSE:
            self.putx(s, [3, ['Selector: %02X' % s[2]]])
        elif op == OPCODE_READ:
            self.putx(s, [3, ['Zone: {}, Length: {}'.format(ZONES[s[2] & 0x03],
                '32 bytes' if s[2] & 0x90 else '4 bytes')]])
        elif op == OPCODE_WRITE:
            self.putx(s, [3, ['Zone: {}, Encrypted: {}, Length: {}'.format(ZONES[s[2] & 0x03],
                'Yes' if s[2] & 0x40 else 'No', '32 bytes' if s[2] & 0x90 else '4 bytes')]])
        else:
            self.putx(s, [3, ['Param1: %02X' % s[2]]])

    def put_param2(self, s):
        op = self.opcode
        if op == OPCODE_DERIVE_KEY:
            self.puty(s, [4, ['TargetKey: {:02x} {:02x}'.format(s[1][2], s[0][2])]])
        elif op in (OPCODE_COUNTER, OPCODE_ECDH, OPCODE_GEN_KEY, OPCODE_PRIVWRITE, \
                    OPCODE_SIGN, OPCODE_VERIFY):
            self.puty(s, [4, ['KeyID: {:02x} {:02x}'.format(s[1][2], s[0][2])]])
        elif op in (OPCODE_NONCE, OPCODE_PAUSE, OPCODE_RANDOM):
            self.puty(s, [4, ['Zero: {:02x} {:02x}'.format(s[1][2], s[0][2])]])
        elif op in (OPCODE_HMAC, OPCODE_MAC, OPCODE_CHECK_MAC, OPCODE_GEN_DIG):
            self.puty(s, [4, ['SlotID: {:02x} {:02x}'.format(s[1][2], s[0][2])]])
        elif op == OPCODE_LOCK:
            self.puty(s, [4, ['Summary: {:02x} {:02x}'.format(s[1][2], s[0][2])]])
        elif op in (OPCODE_READ, OPCODE_WRITE):
            self.puty(s, [4, ['Address: {:02x} {:02x}'.format(s[1][2], s[0][2])]])
        elif op == OPCODE_UPDATE_EXTRA:
            self.puty(s, [4, ['NewValue: {:02x}'.format(s[0][2])]])
        else:
            self.puty(s, [4, ['-']])

    def put_data(self, s):
        if len(s) == 0:
            return
        op = self.opcode
        if op == OPCODE_CHECK_MAC:
            self.putz(s[0][0], s[31][1], [5, ['ClientChal: %s' % ' '.join(format(i[2], '02x') for i in s[0:32])]])
            self.putz(s[32][0], s[63][1], [5, ['ClientResp: %s' % ' '.join(format(i[2], '02x') for i in s[32:64])]])
            self.putz(s[64][0], s[76][1], [5, ['OtherData: %s' % ' '.join(format(i[2], '02x') for i in s[64:77])]])
        elif op == OPCODE_DERIVE_KEY:
            self.putz(s[0][0], s[31][1], [5, ['MAC: %s' % ' '.join(format(i[2], '02x') for i in s)]])
        elif op == OPCODE_ECDH:
            self.putz(s[0][0], s[31][1], [5, ['Pub X: %s' % ' '.join(format(i[2], '02x') for i in s[0:32])]])
            self.putz(s[32][0], s[63][1], [5, ['Pub Y: %s' % ' '.join(format(i[2], '02x') for i in s[32:64])]])
        elif op in (OPCODE_GEN_DIG, OPCODE_GEN_KEY):
            self.putz(s[0][0], s[3][1], [5, ['OtherData: %s' % ' '.join(format(i[2], '02x') for i in s)]])
        elif op == OPCODE_MAC:
            self.putz(s[0][0], s[31][1], [5, ['Challenge: %s' % ' '.join(format(i[2], '02x') for i in s)]])
        elif op == OPCODE_PRIVWRITE:
            if len(s) > 36: # Key + MAC.
                self.putz(s[0][0], s[-35][1], [5, ['Value: %s' % ' '.join(format(i[2], '02x') for i in s)]])
                self.putz(s[-32][0], s[-1][1], [5, ['MAC: %s' % ' '.join(format(i[2], '02x') for i in s)]])
            else: # Just value.
                self.putz(s[0][0], s[-1][1], [5, ['Value: %s' % ' '.join(format(i[2], '02x') for i in s)]])
        elif op == OPCODE_VERIFY:
            if len(s) >= 64: # ECDSA components (always present)
                self.putz(s[0][0], s[31][1], [5, ['ECDSA R: %s' % ' '.join(format(i[2], '02x') for i in s[0:32])]])
                self.putz(s[32][0], s[63][1], [5, ['ECDSA S: %s' % ' '.join(format(i[2], '02x') for i in s[32:64])]])
            if len(s) == 83: # OtherData (follow ECDSA components in validate / invalidate mode)
                self.putz(s[64][0], s[82][1], [5, ['OtherData: %s' % ' '.join(format(i[2], '02x') for i in s[64:83])]])
            if len(s) == 128: # Public key components (follow ECDSA components in external mode)
                self.putz(s[64][0], s[95][1], [5, ['Pub X: %s' % ' '.join(format(i[2], '02x') for i in s[64:96])]])
                self.putz(s[96][0], s[127][1], [5, ['Pub Y: %s' % ' '.join(format(i[2], '02x') for i in s[96:128])]])
        elif op == OPCODE_WRITE:
            if len(s) > 32: # Value + MAC.
                self.putz(s[0][0], s[-31][1], [5, ['Value: %s' % ' '.join(format(i[2], '02x') for i in s)]])
                self.putz(s[-32][0], s[-1][1], [5, ['MAC: %s' % ' '.join(format(i[2], '02x') for i in s)]])
            else: # Just value.
                self.putz(s[0][0], s[-1][1], [5, ['Value: %s' % ' '.join(format(i[2], '02x') for i in s)]])
        else:
            self.putz(s[0][0], s[-1][1], [5, ['Data: %s' % ' '.join(format(i[2], '02x') for i in s)]])

    def put_crc(self, s):
        self.puty(s, [6, ['CRC: {:02X} {:02X}'.format(s[0][2], s[1][2])]])

    def put_status(self, ss, es, status):
        self.putz(ss, es, [7, ['Status: %s' % STATUS[status]]])

    def put_warning(self, ss, es, msg):
        self.putz(ss, es, [8, ['Warning: %s' % msg]])

    def decode(self, ss, es, data):
        cmd, databyte = data
        # State machine.
        if self.state == 'IDLE':
            # Wait for an I²C START condition.
            if cmd != 'START':
                return
            self.state = 'GET SLAVE ADDR'
            self.ss_block = ss
        elif self.state == 'GET SLAVE ADDR':
            # Wait for an address read/write operation.
            if cmd == 'ADDRESS READ':
                self.state = 'READ REGS'
            elif cmd == 'ADDRESS WRITE':
                self.state = 'WRITE REGS'
        elif self.state == 'READ REGS':
            if cmd == 'DATA READ':
                self.bytes.append([ss, es, databyte])
            elif cmd == 'STOP':
                self.es_block = es
                # Reset the opcode before received data, as this causes
                # responses to be displayed incorrectly.
                self.opcode = -1
                if len(self.bytes) > 0:
                    self.output_rx_bytes()
                self.waddr = -1
                self.bytes = []
                self.state = 'IDLE'
        elif self.state == 'WRITE REGS':
            if cmd == 'DATA WRITE':
                self.bytes.append([ss, es, databyte])
            elif cmd == 'STOP':
                self.es_block = es
                self.output_tx_bytes()
                self.bytes = []
                self.state = 'IDLE'
